#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock
from oslo_utils import uuidutils
import tempfile
import threading

from esi_leap.common import exception
from esi_leap.common import statuses
from esi_leap.objects import fields as obj_fields
from esi_leap.objects import lease as lease_obj
from esi_leap.objects import offer as offer_obj
from esi_leap.resource_objects.test_node import TestNode
from esi_leap.tests import base


class TestLeaseObject(base.DBTestCase):

    def setUp(self):
        super(TestLeaseObject, self).setUp()

        self.start_time = datetime.datetime(2016, 7, 16, 19, 20, 30)
        self.test_offer = offer_obj.Offer(
            id=27,
            uuid=uuidutils.generate_uuid(),
            project_id='01d4e6a72f5c408813e02f664cc8c83e',
            resource_type='dummy_node',
            resource_uuid='1718',
            start_time=self.start_time,
            end_time=self.start_time + datetime.timedelta(days=100),
            status=statuses.AVAILABLE,
            properties={'floor_price': 3},
        )
        self.test_parent_lease = lease_obj.Lease(
            uuid=uuidutils.generate_uuid(),
            status=statuses.ACTIVE,
        )
        self.test_parent_lease_expired = lease_obj.Lease(
            uuid=uuidutils.generate_uuid(),
            status=statuses.EXPIRED,
        )
        self.test_lease_dict = {
            'id': 28,
            'name': 'lease',
            'uuid': uuidutils.generate_uuid(),
            'project_id': 'le55ee',
            'owner_id': '0wn3r',
            'start_time': self.start_time + datetime.timedelta(days=5),
            'end_time': self.start_time + datetime.timedelta(days=10),
            'fulfill_time': self.start_time + datetime.timedelta(days=5),
            'expire_time': self.start_time + datetime.timedelta(days=10),
            'status': statuses.CREATED,
            'properties': {},
            'resource_type': 'dummy_node',
            'resource_uuid': '1718',
            'purpose': 'test_purpose',
            'offer_uuid': None,
            'parent_lease_uuid': None,
            'created_at': None,
            'updated_at': None
        }
        self.test_lease_offer_dict = self.test_lease_dict.copy()
        self.test_lease_offer_dict['offer_uuid'] = self.test_offer.uuid
        self.test_lease_parent_lease_dict = self.test_lease_dict.copy()
        self.test_lease_parent_lease_dict['parent_lease_uuid'] = (
            'parent-lease-uuid')
        self.test_lease_create_dict = {
            'name': 'lease_create',
            'project_id': 'le55ee',
            'owner_id': '0wn3r',
            'start_time': self.start_time + datetime.timedelta(days=5),
            'end_time': self.start_time + datetime.timedelta(days=10),
            'resource_type': 'dummy_node',
            'resource_uuid': '1718',
            'purpose': 'test_purpose',
        }
        self.test_lease_create_offer_dict = self.test_lease_create_dict.copy()
        self.test_lease_create_offer_dict['offer_uuid'] = self.test_offer.uuid
        self.test_lease_create_parent_lease_dict = (
            self.test_lease_create_dict.copy())
        self.test_lease_create_parent_lease_dict['parent_lease_uuid'] = (
            'parent-lease-uuid')

        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

    def test_get(self):
        lease_uuid = self.test_lease_dict['uuid']
        with mock.patch.object(self.db_api, 'lease_get_by_uuid',
                               autospec=True) as mock_lease_get_by_uuid:
            mock_lease_get_by_uuid.return_value = self.test_lease_dict

            lease = lease_obj.Lease.get(lease_uuid, self.context)

            mock_lease_get_by_uuid.assert_called_once_with(lease_uuid)
            self.assertEqual(self.context, lease._context)

    def test_get_all(self):
        with mock.patch.object(
                self.db_api, 'lease_get_all', autospec=True
        ) as mock_lease_get_all:
            mock_lease_get_all.return_value = [self.test_lease_dict,
                                               self.test_lease_offer_dict]

            leases = lease_obj.Lease.get_all({}, self.context)

            mock_lease_get_all.assert_called_once_with({})
            self.assertEqual(len(leases), 2)
            self.assertIsInstance(leases[0], lease_obj.Lease)
            self.assertEqual(self.context, leases[0]._context)

    @mock.patch('esi_leap.db.sqlalchemy.api.resource_verify_availability')
    @mock.patch('esi_leap.db.sqlalchemy.api.offer_verify_availability')
    @mock.patch('esi_leap.objects.offer.Offer.get')
    @mock.patch('esi_leap.db.sqlalchemy.api.lease_create')
    def test_create(self, mock_lc, mock_og, mock_ova, mock_rva):
        lease = lease_obj.Lease(self.context, **self.test_lease_create_dict)
        mock_lc.return_value = self.test_lease_dict

        lease.create()

        mock_lc.assert_called_once_with(self.test_lease_create_dict)
        mock_og.assert_not_called
        mock_ova.assert_not_called
        mock_rva.assert_called_once_with(lease.resource_type,
                                         lease.resource_uuid,
                                         lease.start_time,
                                         lease.end_time)

    @mock.patch('esi_leap.db.sqlalchemy.api.resource_verify_availability')
    @mock.patch('esi_leap.db.sqlalchemy.api.offer_verify_availability')
    @mock.patch('esi_leap.objects.offer.Offer.get')
    @mock.patch('esi_leap.db.sqlalchemy.api.lease_create')
    def test_create_with_offer(self, mock_lc, mock_og, mock_ova, mock_rva):
        lease = lease_obj.Lease(self.context,
                                **self.test_lease_create_offer_dict)
        mock_lc.return_value = self.test_lease_offer_dict
        mock_og.return_value = self.test_offer

        lease.create()

        mock_lc.assert_called_once_with(self.test_lease_create_offer_dict)
        mock_og.assert_called_once_with(lease.offer_uuid)
        mock_ova.assert_called_once_with(self.test_offer,
                                         lease.start_time,
                                         lease.end_time)
        mock_rva.assert_not_called

    @mock.patch('esi_leap.db.sqlalchemy.api.lease_verify_child_availability')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.db.sqlalchemy.api.lease_create')
    def test_create_with_parent_lease(self, mock_lc, mock_lg, mock_lvca):
        lease = lease_obj.Lease(self.context,
                                **self.test_lease_create_parent_lease_dict)
        mock_lc.return_value = self.test_lease_offer_dict
        mock_lg.return_value = self.test_parent_lease

        lease.create()

        mock_lc.assert_called_once_with(
            self.test_lease_create_parent_lease_dict)
        mock_lg.assert_called_once_with('parent-lease-uuid')
        mock_lvca.assert_called_once_with(self.test_parent_lease,
                                          lease.start_time,
                                          lease.end_time)

    @mock.patch('esi_leap.db.sqlalchemy.api.lease_verify_child_availability')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.db.sqlalchemy.api.lease_create')
    def test_create_with_parent_lease_expired(self, mock_lc, mock_lg,
                                              mock_lvca):
        lease = lease_obj.Lease(self.context,
                                **self.test_lease_create_parent_lease_dict)
        mock_lg.return_value = self.test_parent_lease_expired

        self.assertRaises(exception.LeaseNotActive, lease.create)

        mock_lc.assert_not_called()
        mock_lg.assert_called_once_with('parent-lease-uuid')
        mock_lvca.assert_not_called()

    def test_create_invalid_time(self):
        bad_lease = {
            'id': 30,
            'name': 'bad-lease',
            'uuid': '534653c9-880d-4c2d-6d6d-44444444444',
            'project_id': 'le55ee_2',
            'owner_id': 'ownerid',
            'resource_type': 'dummy_node',
            'resource_uuid': '1111',
            'start_time': self.start_time + datetime.timedelta(days=30),
            'end_time': self.start_time + datetime.timedelta(days=20),
            'fulfill_time': self.start_time + datetime.timedelta(days=35),
            'expire_time': self.start_time + datetime.timedelta(days=40),
        }

        lease = lease_obj.Lease(self.context, **bad_lease)

        self.assertRaises(exception.InvalidTimeRange, lease.create)

    def test_create_concurrent_offer_conflict(self):
        lease = lease_obj.Lease(self.context,
                                **self.test_lease_create_offer_dict)

        lease2 = lease_obj.Lease(self.context,
                                 **self.test_lease_create_offer_dict)

        lease2.id = 28

        with mock.patch.object(self.db_api, 'lease_create',
                               autospec=True) as mock_lease_create:
            with mock.patch.object(offer_obj.Offer, 'get',
                                   autospec=True) as mock_offer_get:
                with mock.patch.object(self.db_api,
                                       'offer_verify_availability',
                                       autospec=True) as mock_ovca:

                    mock_offer_get.return_value = self.test_offer

                    def update_mock(updates):
                        mock_ovca.side_effect = Exception('bad')

                    mock_lease_create.side_effect = update_mock

                    thread = threading.Thread(target=lease.create)
                    thread2 = threading.Thread(target=lease2.create)

                    thread.start()
                    thread2.start()

                    thread.join()
                    thread2.join()

                    assert mock_offer_get.call_count == 2
                    assert mock_ovca.call_count == 2
                    mock_lease_create.assert_called_once()

    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_fulfill(self, mock_notify,
                     mock_save, mock_set_lease, mock_ro):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node

        lease.fulfill()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'fulfill',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'fulfill',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END,
                                      mock.ANY, node=mock.ANY)])
        mock_ro.assert_called_once()
        mock_set_lease.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.ACTIVE)

    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_fulfill_error(self, mock_notify, mock_save,
                           mock_set_lease, mock_ro):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_set_lease.side_effect = Exception('bad')

        lease.fulfill()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'fulfill',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'fulfill',
                                      obj_fields.NotificationLevel.ERROR,
                                      obj_fields.NotificationStatus.ERROR,
                                      mock.ANY, node=mock.ANY)])
        mock_ro.assert_called_once()
        mock_set_lease.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.WAIT_FULFILL)

    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_cancel(self, mock_notify, mock_save,
                    mock_rl, mock_glu, mock_ro,
                    mock_lg, mock_sl):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = lease.uuid

        lease.cancel()
        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END,
                                      mock.ANY, node=mock.ANY)])
        mock_sl.assert_not_called()
        mock_lg.assert_not_called()
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.DELETED)

    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_cancel_error(self, mock_notify, mock_save,
                          mock_rl, mock_glu,
                          mock_ro, mock_lg, mock_sl):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = lease.uuid
        mock_rl.side_effect = Exception('bad')

        lease.cancel()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.ERROR,
                                      obj_fields.NotificationStatus.ERROR,
                                      mock.ANY, node=mock.ANY)])
        mock_sl.assert_not_called()
        mock_lg.assert_not_called()
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.WAIT_CANCEL)

    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_cancel_with_parent(self, mock_notify, mock_save,
                                mock_rl, mock_glu,
                                mock_ro, mock_lg, mock_sl):
        lease = lease_obj.Lease(self.context,
                                **self.test_lease_parent_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = lease.uuid

        lease.cancel()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END,
                                      mock.ANY, node=mock.ANY)])
        mock_sl.assert_called_once()
        mock_lg.assert_called_once()
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.DELETED)

    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_cancel_no_expire(self, mock_notify, mock_save,
                              mock_rl, mock_glu,
                              mock_ro):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = 'some-other-lease-uuid'

        lease.cancel()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END,
                                      mock.ANY, node=mock.ANY)])
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_not_called()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.DELETED)

    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_expire(self, mock_notify, mock_save, mock_rl, mock_glu, mock_ro,
                    mock_lg, mock_sl):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = lease.uuid

        lease.expire()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END,
                                      mock.ANY, node=mock.ANY)])
        mock_sl.assert_not_called()
        mock_lg.assert_not_called()
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.EXPIRED)

    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_expire_error(self, mock_notify, mock_save, mock_rl, mock_glu,
                          mock_ro, mock_lg, mock_sl):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = lease.uuid
        mock_rl.side_effect = Exception('bad')

        lease.expire()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.ERROR,
                                      obj_fields.NotificationStatus.ERROR,
                                      mock.ANY, node=mock.ANY)])
        mock_sl.assert_not_called()
        mock_lg.assert_not_called()
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.WAIT_EXPIRE)

    @mock.patch('esi_leap.resource_objects.test_node.TestNode.set_lease')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_expire_with_parent(self, mock_notify, mock_save, mock_rl,
                                mock_glu, mock_ro, mock_lg, mock_sl):
        lease = lease_obj.Lease(self.context,
                                **self.test_lease_parent_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = lease.uuid

        lease.expire()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END,
                                      mock.ANY, node=mock.ANY)])
        mock_sl.assert_called_once()
        mock_lg.assert_called_once()
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_called_once()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.EXPIRED)

    @mock.patch('esi_leap.objects.lease.Lease.resource_object')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.test_node.TestNode.remove_lease')
    @mock.patch('esi_leap.objects.lease.Lease.save')
    @mock.patch('esi_leap.common.notification_utils'
                '._emit_notification')
    def test_expire_no_expire(self, mock_notify, mock_save, mock_rl,
                              mock_glu, mock_ro):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        test_node = TestNode('test-node', '12345')

        mock_ro.return_value = test_node
        mock_glu.return_value = 'some-other-lease-uuid'

        lease.expire()

        mock_notify.assert_has_calls([mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.START,
                                      mock.ANY, node=mock.ANY),
                                      mock.call(mock.ANY, mock.ANY, 'delete',
                                      obj_fields.NotificationLevel.INFO,
                                      obj_fields.NotificationStatus.END,
                                      mock.ANY, node=mock.ANY)])
        mock_ro.assert_called_once()
        mock_glu.assert_called_once()
        mock_rl.assert_not_called()
        mock_save.assert_called_once()
        self.assertEqual(lease.status, statuses.EXPIRED)

    def test_destroy(self):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        with mock.patch.object(self.db_api, 'lease_destroy',
                               autospec=True) as mock_lease_cancel:
            lease.destroy()
            mock_lease_cancel.assert_called_once_with(lease.uuid)

    def test_save(self):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        new_status = statuses.ACTIVE
        updated_at = datetime.datetime(2006, 12, 11, 0, 0)
        with mock.patch.object(self.db_api, 'lease_update',
                               autospec=True) as mock_lease_update:
            updated_lease = self.test_lease_dict.copy()
            updated_lease['status'] = new_status
            updated_lease['updated_at'] = updated_at
            mock_lease_update.return_value = updated_lease

            lease.status = new_status
            lease.save(self.context)

            updated_values = self.test_lease_dict.copy()
            updated_values['status'] = new_status
            mock_lease_update.assert_called_once_with(lease.uuid,
                                                      updated_values)
            self.assertEqual(self.context, lease._context)
            self.assertEqual(updated_at, lease.updated_at)

    @mock.patch('esi_leap.objects.lease.get_resource_object')
    def test_resource_object(self, mock_gro):
        lease = lease_obj.Lease(self.context, **self.test_lease_dict)
        lease.resource_object()
        mock_gro.assert_called_once_with(lease.resource_type,
                                         lease.resource_uuid)


class TestLeaseCRUDPayloads(base.DBTestCase):

    def setUp(self):
        super(TestLeaseCRUDPayloads, self).setUp()
        self.lease = lease_obj.Lease(
            id='12345',
            name='test_lease',
            start_time=datetime.datetime(2016, 7, 16, 19, 20, 30),
            end_time=datetime.datetime(2016, 8, 16, 19, 20, 30),
            fulfill_time=datetime.datetime(2016, 7, 16, 19, 21, 30),
            expire_time=datetime.datetime(2016, 8, 16, 19, 21, 30),
            uuid='13921c8d-ce11-4b6d-99ed-10e19d184e5f',
            resource_type='test_node',
            resource_uuid='111',
            project_id='lesseeid',
            owner_id='ownerid',
            parent_lease_uuid=None,
            offer_uuid=None,
            properties=None,
            status='created',
            purpose=None,
        )
        self.node = TestNode('test-node', '12345')

    def test_lease_crud_payload(self):
        payload = lease_obj.LeaseCRUDPayload(self.lease, self.node)
        self.assertEqual(self.lease.id, payload.id)
        self.assertEqual(self.lease.name, payload.name)
        self.assertEqual(self.lease.start_time, payload.start_time)
        self.assertEqual(self.lease.end_time, payload.end_time)
        self.assertEqual(self.lease.fulfill_time, payload.fulfill_time)
        self.assertEqual(self.lease.expire_time, payload.expire_time)
        self.assertEqual(self.lease.uuid, payload.uuid)
        self.assertEqual(self.lease.resource_type, payload.resource_type)
        self.assertEqual(self.lease.resource_uuid, payload.resource_uuid)
        self.assertEqual(self.lease.project_id, payload.project_id)
        self.assertEqual(self.lease.owner_id, payload.owner_id)
        self.assertEqual(self.lease.parent_lease_uuid,
                         payload.parent_lease_uuid)
        self.assertEqual(self.lease.offer_uuid, payload.offer_uuid)
        self.assertEqual(self.lease.properties, payload.properties)
        self.assertEqual(self.lease.status, payload.status)
        self.assertEqual(self.lease.purpose, payload.purpose)
        self.assertEqual(self.node.node_name, payload.node_name)
        self.assertEqual(self.node._uuid, payload.node_uuid)
        self.assertEqual(self.node.node_power_state,
                         payload.node_power_state)
        self.assertEqual(self.node.node_provision_state,
                         payload.node_provision_state),
        self.assertEqual(self.node.node_properties,
                         payload.node_properties)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from esi_leap.common import statuses
import esi_leap.conf
from esi_leap.manager import utils
from esi_leap.objects import lease as lease_obj
from esi_leap.objects import offer as offer_obj
from oslo_context import context as ctx
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import service
from oslo_utils import timeutils

CONF = esi_leap.conf.CONF
EVENT_INTERVAL = 60
LOG = logging.getLogger(__name__)


class ManagerService(service.Service):
    def __init__(self):
        super(ManagerService, self).__init__()
        LOG.info('Creating esi-leap manager RPC server')
        self._server = messaging.get_rpc_server(
            target=utils.get_target(),
            transport=messaging.get_rpc_transport(CONF),
            endpoints=[ManagerEndpoint()],
            executor='eventlet',
        )
        self._context = ctx.RequestContext(
            auth_token=None,
            project_id=None,
            overwrite=False)

    def start(self):
        super(ManagerService, self).start()
        LOG.info('Starting esi-leap manager RPC server')
        self.tg.add_thread(self._server.start)
        LOG.info('Starting _fulfill_leases periodic job')
        self.tg.add_timer(EVENT_INTERVAL, self._fulfill_leases)
        LOG.info('Starting _expire_leases periodic job')
        self.tg.add_timer(EVENT_INTERVAL, self._expire_leases)
        LOG.info('Starting _cancel_leases periodic job')
        self.tg.add_timer(EVENT_INTERVAL, self._cancel_leases)
        LOG.info('Starting _expire_offers periodic job')
        self.tg.add_timer(EVENT_INTERVAL, self._expire_offers)

    def stop(self):
        super(ManagerService, self).stop()
        LOG.info('Shutting down esi-leap manager RPC server')
        self._server.stop()

    def _fulfill_leases(self):
        LOG.info('Checking for leases to fulfill')
        leases = lease_obj.Lease.get_all(
            {'status': [statuses.CREATED, statuses.WAIT_FULFILL]},
            self._context)
        now = timeutils.utcnow()
        for lease in leases:
            if lease.start_time <= now and now <= lease.end_time:
                try:
                    LOG.info('Fulfilling lease %s', lease.uuid)
                    lease.fulfill(self._context)
                except Exception as e:
                    LOG.info('Error fulfilling lease: %s: %s' %
                             (type(e).__name__, e))
                    LOG.info('Setting lease status to ERROR')
                    lease.status = statuses.ERROR
                    lease.save()

    def _expire_leases(self):
        LOG.info('Checking for expiring leases')
        leases = lease_obj.Lease.get_all(
            {'status': [statuses.ACTIVE, statuses.CREATED,
                        statuses.WAIT_EXPIRE, statuses.WAIT_FULFILL]},
            self._context)
        now = timeutils.utcnow()
        for lease in leases:
            if lease.end_time <= now:
                try:
                    LOG.info('Expiring lease %s', lease.uuid)
                    lease.expire(self._context)
                except Exception as e:
                    LOG.info('Error expiring lease: %s: %s' %
                             (type(e).__name__, e))
                    LOG.info('Setting lease status to ERROR')
                    lease.status = statuses.ERROR
                    lease.save()

    def _cancel_leases(self):
        LOG.info('Checking for leases to cancel')
        leases = lease_obj.Lease.get_all(
            {'status': [statuses.WAIT_CANCEL]}, self._context)
        for lease in leases:
            try:
                LOG.info('Cancelling lease %s', lease.uuid)
                lease.cancel()
            except Exception as e:
                LOG.info('Error cancelling lease: %s: %s' %
                         (type(e).__name__, e))
                LOG.info('Setting lease status to ERROR')
                lease.status = statuses.ERROR
                lease.save()

    def _expire_offers(self):
        LOG.info('Checking for expiring offers')
        offers = offer_obj.Offer.get_all({'status': statuses.OFFER_CAN_DELETE},
                                         self._context)

        for offer in offers:
            if offer.end_time and offer.end_time <= timeutils.utcnow():
                try:
                    LOG.info('Expiring offer %s for %s %s',
                             offer.uuid, offer.resource_type,
                             offer.resource_uuid)
                    offer.expire(self._context)
                except Exception as e:
                    LOG.info('Error expiring offer: %s: %s' %
                             (type(e).__name__, e))
                    offer.status = statuses.ERROR
                    offer.save()


class ManagerEndpoint(object):
    target = utils.get_target()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from esi_leap.common import ironic
from esi_leap.tests import base


class FakeNode(object):
    def __init__(self):
        self.uuid = 'uuid'
        self.name = 'name'
        self.resource_class = 'baremetal'


class IronicTestCase(base.TestCase):

    @mock.patch.object(ironic, 'get_ironic_client', autospec=True)
    def test_get_node(self, mock_ironic):
        fake_node = FakeNode()
        mock_ironic.return_value.node.get.return_value = fake_node
        node = ironic.get_node('12345')

        self.assertEqual(fake_node, node)

    @mock.patch.object(ironic, 'get_ironic_client', autospec=True)
    def test_get_node_list(self, mock_ironic):
        fake_node = FakeNode()
        node_list = [fake_node]
        node = ironic.get_node('uuid', node_list)

        self.assertEqual(fake_node, node)

    @mock.patch.object(ironic, 'get_ironic_client', autospec=True)
    def test_get_node_list_no_match(self, mock_ironic):
        fake_node = FakeNode()
        node_list = [fake_node]
        node = ironic.get_node('uuid2', node_list)

        self.assertEqual(None, node)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

import mock
from oslo_utils import uuidutils

from esi_leap.resource_objects.base import ResourceObjectInterface
from esi_leap.tests import base


def bind_abstract_methods(cls):
    # binds all undefined abstract methods to stub functions that do nothing
    def wrapper(*args, **kwargs):
        for method in cls.__abstractmethods__:
            setattr(cls, method, lambda: None)
        methods = set()
        for name, value in cls.__dict__.items():
            if getattr(value, "__isabstractmethod__", False):
                methods.add(name)
        cls.__abstractmethods__ = frozenset(methods)
        return cls(*args, **kwargs)
    return wrapper


@bind_abstract_methods
class ResourceObjectStub(ResourceObjectInterface):
    def __init__(self, uuid):
        self._uuid = uuid

    def get_uuid(self):
        return self._uuid


class TestResourceObjectInterface(base.TestCase):

    @mock.patch('esi_leap.db.sqlalchemy.api.resource_verify_availability')
    def test_verify_availability_for_offer(self, mock_rva):
        start = datetime.datetime(2016, 7, 16, 19, 20, 30)
        end = start + datetime.timedelta(days=10)
        fake_uuid = uuidutils.generate_uuid()
        test_node = ResourceObjectStub(fake_uuid)

        test_node.verify_availability(start, end)
        mock_rva.assert_called_once_with('base', fake_uuid, start, end)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import esi_leap.conf

_opts = [
    ('DEFAULT', esi_leap.conf.netconf.opts),
    ('api', esi_leap.conf.api.opts),
    ('dummy_node', esi_leap.conf.dummy_node.opts),
    ('ironic', esi_leap.conf.ironic.list_opts()),
    ('keystone', esi_leap.conf.keystone.list_opts()),
    ('pecan', esi_leap.conf.pecan.opts),
    ('notification', esi_leap.conf.notification.opts),
]


def list_opts():
    return _opts



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import mock

from esi_leap.objects import event as event_obj
from esi_leap.tests import base


class TestEventObject(base.DBTestCase):

    def setUp(self):
        super(TestEventObject, self).setUp()

        event_time = datetime.now()
        self.test_event_create_dict = {
            'event_type': 'fake:event',
            'event_time': event_time,
            'object_type': 'lease',
            'object_uuid': '11111',
            'resource_type': 'dummy_node',
            'resource_uuid': '22222',
            'lessee_id': '33333',
            'owner_id': '44444',
        }
        self.test_event_dict = {
            'id': 1,
            'event_type': 'fake:event',
            'event_time': event_time,
            'object_type': 'lease',
            'object_uuid': '11111',
            'resource_type': 'dummy_node',
            'resource_uuid': '22222',
            'lessee_id': '33333',
            'owner_id': '44444',
            'created_at': event_time,
            'updated_at': None,
        }

    @mock.patch('esi_leap.db.sqlalchemy.api.event_get_all')
    def test_get_all(self, mock_ega):
        event_obj.Event.get_all({}, self.context)
        mock_ega.assert_called_once_with({})

    @mock.patch('esi_leap.db.sqlalchemy.api.event_create')
    def test_create(self, mock_ec):
        mock_ec.return_value = self.test_event_dict
        event = event_obj.Event(self.context,
                                **self.test_event_create_dict)
        event.create()
        mock_ec.assert_called_once_with(self.test_event_create_dict)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import models
from oslo_db.sqlalchemy import types as db_types

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import orm
from sqlalchemy import Column, DateTime, ForeignKey
from sqlalchemy import Index, Integer, String

from esi_leap.common import statuses


@compiles(DateTime, 'mysql')
def compile_datetime_mysql(type_, compiler, **kw):
    return 'DATETIME(6)'


class ESILEAPBase(models.TimestampMixin, models.ModelBase):

    metadata = None

    def to_dict(self):
        d = {}
        for c in self.__table__.columns:
            d[c.name] = self[c.name]
        return d


Base = declarative_base(cls=ESILEAPBase)


class Offer(Base):
    """Represents a resource that is offered."""

    __tablename__ = 'offers'
    __table_args__ = (
        Index('offer_uuid_idx', 'uuid'),
        Index('offer_project_id_idx', 'project_id'),
        Index('offer_resource_idx', 'resource_type', 'resource_uuid'),
        Index('offer_status_idx', 'status'),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    uuid = Column(String(36), nullable=False, unique=True)
    name = Column(String(35), nullable=True, unique=False)
    project_id = Column(String(255), nullable=False)
    lessee_id = Column(String(255), nullable=True)
    resource_type = Column(String(36), nullable=False)
    resource_uuid = Column(String(36), nullable=False)
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    status = Column(String(15), nullable=False, default=statuses.AVAILABLE)
    properties = Column(db_types.JsonEncodedDict, nullable=True)
    parent_lease_uuid = Column(String(36),
                               ForeignKey('leases.uuid'),
                               nullable=True)
    parent_lease = orm.relationship(
        'Lease',
        foreign_keys=[parent_lease_uuid],
    )


class Lease(Base):
    """Represents a lease."""

    __tablename__ = 'leases'
    __table_args__ = (
        Index('lease_uuid_idx', 'uuid'),
        Index('lease_project_id_idx', 'project_id'),
        Index('lease_owner_id_idx', 'owner_id'),
        Index('lease_status_idx', 'status'),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    uuid = Column(String(36), nullable=False, unique=True)
    name = Column(String(35), nullable=True, unique=False)
    project_id = Column(String(255), nullable=False)
    owner_id = Column(String(255), nullable=False)
    purpose = Column(String(255), nullable=True)
    resource_type = Column(String(36), nullable=False)
    resource_uuid = Column(String(36), nullable=False)
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    fulfill_time = Column(DateTime)
    expire_time = Column(DateTime)
    status = Column(String(15), nullable=False, default=statuses.CREATED)
    properties = Column(db_types.JsonEncodedDict, nullable=True)
    offer_uuid = Column(String(36),
                        ForeignKey('offers.uuid'),
                        nullable=True)
    parent_lease_uuid = Column(String(36),
                               ForeignKey('leases.uuid'),
                               nullable=True)
    offer = orm.relationship(
        Offer,
        backref=orm.backref('offers'),
        foreign_keys=offer_uuid,
        primaryjoin=offer_uuid == Offer.uuid)
    parent_lease = orm.relationship(
        'Lease',
        backref=orm.backref('child_leases', remote_side=uuid),
    )


class Event(Base):
    """Represents an event."""

    __tablename__ = 'events'
    __table_args__ = (
        Index('event_type_idx', 'event_type'),
        Index('event_lessee_id_idx', 'lessee_id'),
        Index('event_owner_id_idx', 'owner_id'),
        Index('event_resource_idx', 'resource_type', 'resource_uuid'),
        Index('event_time_idx', 'event_time'),
    )

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    event_type = Column(String(36), nullable=False)
    event_time = Column(DateTime, nullable=False)
    object_type = Column(String(36), nullable=True)
    object_uuid = Column(String(36), nullable=True)
    resource_type = Column(String(36), nullable=True)
    resource_uuid = Column(String(36), nullable=True)
    lessee_id = Column(String(255), nullable=True)
    owner_id = Column(String(255), nullable=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import options as db_options
from oslo_log import log
from oslo_service import service

from esi_leap.common import rpc
import esi_leap.conf
from esi_leap import objects
from esi_leap import version

CONF = esi_leap.conf.CONF


def prepare_service(argv=None, default_config_files=None):
    argv = [] if argv is None else argv
    log.register_options(CONF)
    CONF(argv[1:],
         project='esi-leap',
         version=version.version_info.release_string(),
         default_config_files=default_config_files)
    db_options.set_defaults(CONF)
    log.setup(CONF, 'esi-leap')
    rpc.init(CONF)
    objects.register_all()


def process_launcher():
    return service.ProcessLauncher(CONF, restart_method='mutate')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock
from oslo_utils import uuidutils
import tempfile
import threading

from esi_leap.common import exception
from esi_leap.common import statuses
from esi_leap.objects import lease
from esi_leap.objects import offer
from esi_leap.tests import base


class TestOfferObject(base.DBTestCase):

    def setUp(self):
        super(TestOfferObject, self).setUp()

        start = datetime.datetime(2016, 7, 16, 19, 20, 30)
        self.test_offer_data = {
            'id': 27,
            'name': 'o',
            'uuid': uuidutils.generate_uuid(),
            'project_id': '0wn5r',
            'lessee_id': None,
            'resource_type': 'dummy_node',
            'resource_uuid': '1718',
            'start_time': start,
            'end_time': start + datetime.timedelta(days=100),
            'status': statuses.AVAILABLE,
            'properties': {'floor_price': 3},
            'parent_lease_uuid': None,
            'created_at': None,
            'updated_at': None
        }
        self.test_offer_create_data = {
            'name': 'o',
            'project_id': '0wn5r',
            'resource_type': 'dummy_node',
            'resource_uuid': '1718',
            'start_time': start,
            'end_time': start + datetime.timedelta(days=100),
            'properties': {'floor_price': 3},
        }
        self.test_offer_create_parent_lease_data = (
            self.test_offer_create_data.copy())
        self.test_offer_create_parent_lease_data['parent_lease_uuid'] = (
            'parent-lease-uuid')
        self.test_parent_lease = lease.Lease(
            uuid=uuidutils.generate_uuid(),
            status=statuses.ACTIVE)
        self.test_parent_lease_expired = lease.Lease(
            uuid=uuidutils.generate_uuid(),
            status=statuses.EXPIRED)

        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_get_by_uuid')
    def test_get(self, mock_offer_get_by_uuid):
        offer_uuid = self.test_offer_data['uuid']
        mock_offer_get_by_uuid.return_value = self.test_offer_data

        o = offer.Offer.get(offer_uuid, self.context)

        mock_offer_get_by_uuid.assert_called_once_with(offer_uuid)
        self.assertEqual(self.context, o._context)

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_get_all')
    def test_get_all(self, mock_offer_get_all):
        mock_offer_get_all.return_value = [self.test_offer_data]

        offers = offer.Offer.get_all({}, self.context)

        mock_offer_get_all.assert_called_once_with({})
        self.assertEqual(len(offers), 1)
        self.assertIsInstance(offers[0], offer.Offer)
        self.assertEqual(self.context, offers[0]._context)

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_get_conflict_times')
    @mock.patch('esi_leap.objects.offer.datetime')
    def test_get_availabilities_offer_in_future(self, mock_datetime,
                                                mock_ogct):
        o = offer.Offer(self.context, **self.test_offer_data)

        # test offer start_time > now
        mock_datetime.datetime.now = mock.Mock(
            return_value=o.start_time + datetime.timedelta(days=-5))

        mock_ogct.return_value = [
            [o.start_time + datetime.timedelta(days=10),
             o.start_time + datetime.timedelta(days=20)],
            [o.start_time + datetime.timedelta(days=20),
             o.start_time + datetime.timedelta(days=30)],
            [o.start_time + datetime.timedelta(days=50),
             o.start_time + datetime.timedelta(days=60)]
        ]
        expect = [
            [o.start_time, o.start_time + datetime.timedelta(days=10)],
            [o.start_time + datetime.timedelta(days=30),
             o.start_time + datetime.timedelta(days=50)],
            [o.start_time + datetime.timedelta(days=60), o.end_time]
        ]
        a = o.get_availabilities()
        self.assertEqual(a, expect)

        mock_ogct.return_value = [[o.start_time, o.end_time]]
        expect = []
        a = o.get_availabilities()
        self.assertEqual(a, expect)

        mock_ogct.return_value = []
        expect = [[o.start_time, o.end_time]]
        a = o.get_availabilities()
        self.assertEqual(a, expect)

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_get_conflict_times')
    @mock.patch('esi_leap.objects.offer.datetime')
    def test_get_availabilities_conflicts_in_future(self, mock_datetime,
                                                    mock_ogct):
        o = offer.Offer(self.context, **self.test_offer_data)

        # test offer start_time <= now < first conflict start_time
        now = o.start_time + datetime.timedelta(days=10)
        mock_datetime.datetime.now = mock.Mock(return_value=now)
        mock_ogct.return_value = []
        expect = [[now, o.end_time]]

        a = o.get_availabilities()
        self.assertEqual(a, expect)
        mock_ogct.return_value = [
            [o.start_time + datetime.timedelta(days=10),
             o.start_time + datetime.timedelta(days=20)],
            [o.start_time + datetime.timedelta(days=20),
             o.start_time + datetime.timedelta(days=30)],
            [o.start_time + datetime.timedelta(days=50),
             o.start_time + datetime.timedelta(days=60)]
        ]

        expect = [
            [o.start_time + datetime.timedelta(days=30),
             o.start_time + datetime.timedelta(days=50)],
            [o.start_time + datetime.timedelta(days=60), o.end_time]
        ]
        a = o.get_availabilities()
        self.assertEqual(a, expect)

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_get_conflict_times')
    @mock.patch('esi_leap.objects.offer.datetime')
    def test_get_availabilities_conflicts_current(self, mock_datetime,
                                                  mock_ogct):
        o = offer.Offer(self.context, **self.test_offer_data)

        # test a conflict is happening right now
        now = o.start_time + datetime.timedelta(days=10)
        mock_datetime.datetime.now = mock.Mock(return_value=now)
        mock_ogct.return_value = [
            [o.start_time, o.start_time + datetime.timedelta(days=4)],
            [o.start_time + datetime.timedelta(days=5),
             o.start_time + datetime.timedelta(days=15)],
            [o.start_time + datetime.timedelta(days=20),
             o.start_time + datetime.timedelta(days=30)],
            [o.start_time + datetime.timedelta(days=50),
             o.start_time + datetime.timedelta(days=60)]
        ]
        expect = [
            [o.start_time + datetime.timedelta(days=15),
             o.start_time + datetime.timedelta(days=20)],
            [o.start_time + datetime.timedelta(days=30),
             o.start_time + datetime.timedelta(days=50)],
            [o.start_time + datetime.timedelta(days=60), o.end_time]
        ]
        a = o.get_availabilities()
        self.assertEqual(a, expect)

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_get_conflict_times')
    @mock.patch('esi_leap.objects.offer.datetime')
    def test_get_availabilities_conflicts_past(self, mock_datetime,
                                               mock_ogct):
        o = offer.Offer(self.context, **self.test_offer_data)

        # test all conflicts happened in the past
        now = o.start_time + datetime.timedelta(days=10)
        mock_datetime.datetime.now = mock.Mock(return_value=now)
        mock_ogct.return_value = [
            [o.start_time, o.start_time + datetime.timedelta(days=4)],
            [o.start_time + datetime.timedelta(days=5),
             o.start_time + datetime.timedelta(days=10)]
        ]
        expect = [[now, o.end_time]]
        a = o.get_availabilities()
        self.assertEqual(a, expect)

    @mock.patch('esi_leap.db.sqlalchemy.api.resource_verify_availability')
    @mock.patch('esi_leap.db.sqlalchemy.api.offer_create')
    def test_create(self, mock_oc, mock_rva):
        o = offer.Offer(self.context, **self.test_offer_create_data)
        mock_oc.return_value = self.test_offer_data

        o.create(self.context)

        mock_rva.assert_called_once_with(o.resource_type,
                                         o.resource_uuid,
                                         o.start_time,
                                         o.end_time)
        mock_oc.assert_called_once_with(self.test_offer_create_data)

    def test_create_invalid_time(self):
        start = self.test_offer_data['start_time']
        bad_offer = {
            'name': 'o',
            'project_id': '0wn5r',
            'resource_type': 'dummy_node',
            'resource_uuid': '1718',
            'start_time': start + datetime.timedelta(days=100),
            'end_time': start,
            'status': statuses.AVAILABLE,
            'properties': {'floor_price': 3},
        }

        o = offer.Offer(self.context, **bad_offer)

        self.assertRaises(exception.InvalidTimeRange, o.create)

    @mock.patch('esi_leap.db.sqlalchemy.api.lease_verify_child_availability')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.db.sqlalchemy.api.offer_create')
    def test_create_with_parent_lease(self, mock_oc, mock_lg, mock_lvca):
        o = offer.Offer(self.context,
                        **self.test_offer_create_parent_lease_data)

        mock_lg.return_value = self.test_parent_lease
        mock_oc.return_value = self.test_offer_data

        o.create(self.context)

        mock_lg.assert_called_once_with('parent-lease-uuid')
        mock_lvca.assert_called_once_with(self.test_parent_lease,
                                          o.start_time,
                                          o.end_time)
        mock_oc.assert_called_once_with(
            self.test_offer_create_parent_lease_data)

    @mock.patch('esi_leap.db.sqlalchemy.api.lease_verify_child_availability')
    @mock.patch('esi_leap.objects.lease.Lease.get')
    @mock.patch('esi_leap.db.sqlalchemy.api.offer_create')
    def test_create_with_parent_lease_expired(self, mock_oc, mock_lg,
                                              mock_lvca):
        o = offer.Offer(self.context,
                        **self.test_offer_create_parent_lease_data)

        mock_lg.return_value = self.test_parent_lease_expired
        mock_oc.return_value = self.test_offer_data

        self.assertRaises(exception.LeaseNotActive, o.create, self.context)

        mock_lg.assert_called_once_with('parent-lease-uuid')
        mock_lvca.assert_not_called()
        mock_oc.assert_not_called()

    @mock.patch('esi_leap.db.sqlalchemy.api.resource_verify_availability')
    @mock.patch('esi_leap.db.sqlalchemy.api.offer_create')
    def test_create_concurrent(self, mock_oc, mock_rva):
        o = offer.Offer(self.context, **self.test_offer_create_data)
        o2 = offer.Offer(self.context, **self.test_offer_create_data)
        o2.id = 28

        def update_mock(updates):
            mock_rva.side_effect = Exception('bad')

        mock_oc.side_effect = update_mock

        thread = threading.Thread(target=o.create)
        thread2 = threading.Thread(target=o2.create)

        thread.start()
        thread2.start()

        thread.join()
        thread2.join()

        self.assertEqual(mock_rva.call_count, 2)
        mock_oc.assert_called_once()

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_destroy')
    def test_destroy(self, mock_offer_destroy):
        o = offer.Offer(self.context, **self.test_offer_data)
        o.destroy()
        mock_offer_destroy.assert_called_once_with(o.uuid)

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_update')
    def test_save(self, mock_offer_update):
        o = offer.Offer(self.context, **self.test_offer_data)
        new_status = statuses.DELETED
        updated_at = datetime.datetime(2006, 12, 11, 0, 0)

        updated_offer = self.test_offer_data.copy()
        updated_offer['status'] = new_status
        updated_offer['updated_at'] = updated_at
        mock_offer_update.return_value = updated_offer

        o.status = new_status
        o.save(self.context)

        updated_values = self.test_offer_data.copy()
        updated_values['status'] = new_status
        mock_offer_update.assert_called_once_with(o.uuid, updated_values)
        self.assertEqual(self.context, o._context)
        self.assertEqual(updated_at, o.updated_at)

    @mock.patch('esi_leap.db.sqlalchemy.api.offer_verify_availability')
    def test_verify_availability(self, mock_ova):
        o = offer.Offer(self.context, **self.test_offer_data)
        o.verify_availability(o.start_time, o.end_time)
        mock_ova.assert_called_once_with(o, o.start_time, o.end_time)

    @mock.patch('esi_leap.objects.offer.get_resource_object')
    def test_resource_object(self, mock_gro):
        o = offer.Offer(self.context, **self.test_offer_data)
        o.resource_object()
        mock_gro.assert_called_once_with(o.resource_type, o.resource_uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from esi_leap.common import exception as e
from esi_leap.common import keystone
from esi_leap.tests import base


class FakeProject(object):
    def __init__(self):
        self.id = 'uuid'
        self.name = 'name'


class KeystoneTestCase(base.TestCase):

    @mock.patch('oslo_utils.uuidutils.is_uuid_like')
    def test_get_project_uuid_from_ident_uuid(self, mock_iul):
        mock_iul.return_value = True

        project_uuid = keystone.get_project_uuid_from_ident('uuid')

        mock_iul.assert_called_once_with('uuid')
        self.assertEqual('uuid', project_uuid)

    @mock.patch.object(keystone, 'get_keystone_client', autospec=True)
    @mock.patch('oslo_utils.uuidutils.is_uuid_like')
    def test_get_project_uuid_from_ident_name(self, mock_iul, mock_keystone):
        mock_iul.return_value = False
        mock_keystone.return_value.projects.list.return_value = [FakeProject()]

        project_uuid = keystone.get_project_uuid_from_ident('name')

        mock_iul.assert_called_once_with('name')
        self.assertEqual('uuid', project_uuid)
        mock_keystone.return_value.projects.list.assert_called_once_with(
            name='name')

    @mock.patch.object(keystone, 'get_keystone_client', autospec=True)
    @mock.patch('oslo_utils.uuidutils.is_uuid_like')
    def test_get_project_uuid_from_ident_name_no_match(self, mock_iul,
                                                       mock_keystone):
        mock_iul.return_value = False
        mock_keystone.return_value.projects.list.return_value = []

        self.assertRaises(e.ProjectNoSuchName,
                          keystone.get_project_uuid_from_ident,
                          'name')

        mock_iul.assert_called_once_with('name')
        mock_keystone.return_value.projects.list.assert_called_once_with(
            name='name')

    @mock.patch.object(keystone, 'get_keystone_client', autospec=True)
    def test_get_project_name_no_list(self, mock_keystone):
        mock_keystone.return_value.projects.get.return_value = FakeProject()

        project_name = keystone.get_project_name('12345')

        self.assertEqual('name', project_name)

    @mock.patch.object(keystone, 'get_keystone_client', autospec=True)
    def test_get_project_name_list(self, mock_keystone):
        project_list = [FakeProject()]
        project_name = keystone.get_project_name('uuid', project_list)

        self.assertEqual('name', project_name)

    @mock.patch.object(keystone, 'get_keystone_client', autospec=True)
    def test_get_project_name_list_no_match(self, mock_keystone):
        project_list = [FakeProject()]
        project_name = keystone.get_project_name('uuid2', project_list)

        self.assertEqual('', project_name)

    @mock.patch.object(keystone, 'get_keystone_client', autospec=True)
    def test_get_project_name_none(self, mock_keystone):
        project_list = [FakeProject()]
        project_name = keystone.get_project_name(None, project_list)

        self.assertEqual('', project_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


opts = [
    cfg.HostAddressOpt('host_ip', default='0.0.0.0'),
    cfg.PortOpt('port', default=7777),
    cfg.IntOpt('max_limit', default=1000),
    cfg.StrOpt('public_endpoint'),
    cfg.IntOpt('api_workers'),
    cfg.BoolOpt('enable_ssl_api', default=False),
    cfg.StrOpt('default_resource_type', default='ironic_node'),
    cfg.IntOpt('max_lease_time', default=21),
    cfg.IntOpt('default_lease_time', default=7),
]


api_group = cfg.OptGroup('api', title='API Options')


def register_opts(conf):
    conf.register_opts(opts, group=api_group)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import pecan
import pecan.testing
import tempfile

from esi_leap.api import app
import esi_leap.conf
from esi_leap.tests import base


CONF = esi_leap.conf.CONF
PATH_PREFIX = '/v1'


class APITestCase(base.DBTestCase):

    def setUp(self):
        super(APITestCase, self).setUp()

        CONF.set_override('auth_enable', False, group='pecan')

        self.app = pecan.testing.load_test_app(dict(app.get_pecan_config()))

        self.patch_context = mock.patch(
            'oslo_context.context.RequestContext.from_environ')
        self.mock_context = self.patch_context.start()
        self.addCleanup(self.patch_context.stop)
        self.mock_context.return_value = self.context

        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

    # borrowed from Ironic
    def get_json(self, path, expect_errors=False, headers=None,
                 extra_environ=None, q=None, path_prefix=PATH_PREFIX,
                 **params):
        """Sends simulated HTTP GET request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value;whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param q: list of queries consisting of: field, value, op, and type
                  keys
        :param path_prefix: prefix of the url path
        :param params: content for wsgi.input of request
        """
        q = q if q is not None else []
        full_path = path_prefix + path
        query_params = {'q.field': [],
                        'q.value': [],
                        'q.op': []}
        for query in q:
            for name in ['field', 'op', 'value']:
                query_params['q.%s' % name].append(query.get(name, ''))
        all_params = {}
        all_params.update(params)
        if q:
            all_params.update(query_params)
        response = self.app.get(full_path,
                                params=all_params,
                                headers=headers,
                                extra_environ=extra_environ,
                                expect_errors=expect_errors)
        if not expect_errors:
            response = response.json
        return response

    # borrowed from Ironic
    def post_json(self, path, params, expect_errors=False, headers=None,
                  extra_environ=None, status=None):
        """Sends simulated HTTP POST request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        return self._request_json(path=path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method='post')

    # borrowed from Ironic
    def delete_json(self, path, expect_errors=False, headers=None,
                    extra_environ=None, status=None):
        """Sends simulated HTTP POST request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        return self._request_json(path=path, params={},
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method='delete')

    # borrowed from Ironic
    def _request_json(self, path, params, expect_errors=False, headers=None,
                      method='post', extra_environ=None, status=None,
                      path_prefix=PATH_PREFIX):
        """Sends simulated HTTP request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param method: Request method type. Appropriate method function call
                       should be used rather than passing attribute in.
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        :param path_prefix: prefix of the url path
        """

        full_path = path_prefix + path
        response = getattr(self.app, '%s_json' % method)(
            str(full_path),
            params=params,
            headers=headers,
            status=status,
            extra_environ=extra_environ,
            expect_errors=expect_errors
        )
        return response



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import http.client as http_client
import mock
from oslo_utils import uuidutils

from esi_leap.common import exception
from esi_leap.common import statuses
from esi_leap.objects import offer
from esi_leap.resource_objects.ironic_node import IronicNode
from esi_leap.resource_objects.test_node import TestNode
from esi_leap.tests.api import base as test_api_base


def _get_offer_response(o, use_datetime=False):
    if use_datetime:
        start = datetime.datetime(2016, 7, 16)
        end = datetime.datetime(2016, 10, 24)
    else:
        start = '2016-07-16T00:00:00'
        end = '2016-10-24T00:00:00'
    if o.resource_type in ['test_node', 'dummy_node']:
        resource_class = 'fake'
    elif o.resource_type == 'ironic_node':
        resource_class = 'baremetal'

    return {
        'resource_type': o.resource_type,
        'resource_uuid': o.resource_uuid,
        'resource_class': resource_class,
        'name': o.name,
        'project_id': o.project_id,
        'start_time': start,
        'end_time': end,
        'status': o.status,
        'availabilities': [],
        'uuid': o.uuid,
        'parent_lease_uuid': None
    }


class TestOffersController(test_api_base.APITestCase):

    def setUp(self):
        super(TestOffersController, self).setUp()

        start = datetime.datetime(2016, 7, 16)
        self.test_offer = offer.Offer(
            resource_type='test_node',
            resource_uuid=uuidutils.generate_uuid(),
            name='test_offer',
            uuid=uuidutils.generate_uuid(),
            status=statuses.AVAILABLE,
            start_time=start,
            end_time=start + datetime.timedelta(days=100),
            project_id=self.context.project_id,
            parent_lease_uuid=None
        )
        self.test_offer_drt = offer.Offer(
            resource_type='ironic_node',
            resource_uuid=uuidutils.generate_uuid(),
            name='test_offer',
            uuid=uuidutils.generate_uuid(),
            status=statuses.AVAILABLE,
            start_time=start,
            end_time=start + datetime.timedelta(days=100),
            project_id=self.context.project_id,
            parent_lease_uuid=None
        )
        self.test_offer_lessee = offer.Offer(
            resource_type='test_node',
            resource_uuid=uuidutils.generate_uuid(),
            name='test_offer',
            uuid=uuidutils.generate_uuid(),
            lessee_id='lessee-uuid',
            status=statuses.AVAILABLE,
            start_time=start,
            end_time=start + datetime.timedelta(days=100),
            project_id=self.context.project_id,
            parent_lease_uuid=None
        )
        self.test_offer_2 = offer.Offer(
            resource_type='test_node',
            resource_uuid=uuidutils.generate_uuid(),
            name='test_offer2',
            uuid=uuidutils.generate_uuid(),
            status=statuses.DELETED,
            start_time=start,
            end_time=start + datetime.timedelta(days=100),
            project_id=self.context.project_id,
            parent_lease_uuid=None
        )
        self.test_offer_with_parent = offer.Offer(
            resource_type='test_node',
            resource_uuid=uuidutils.generate_uuid(),
            name='test_offer',
            uuid=uuidutils.generate_uuid(),
            status=statuses.AVAILABLE,
            start_time=start,
            end_time=start + datetime.timedelta(days=100),
            project_id=self.context.project_id,
            parent_lease_uuid='parent-lease-uuid'
        )

    def test_empty(self):
        data = self.get_json('/offers')
        self.assertEqual([], data['offers'])

    @mock.patch('esi_leap.api.controllers.v1.offer.get_resource_object')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.offer.Offer.create')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    def test_post(self, mock_ogdwai, mock_create, mock_cra,
                  mock_generate_uuid, mock_gro):
        resource = TestNode(self.test_offer.resource_uuid)
        mock_gro.return_value = resource
        mock_generate_uuid.return_value = self.test_offer.uuid
        mock_create.return_value = self.test_offer
        mock_ogdwai.return_value = self.test_offer.to_dict()

        data = {
            'resource_type': self.test_offer.resource_type,
            'resource_uuid': self.test_offer.resource_uuid,
            'name': self.test_offer.name,
            'start_time': '2016-07-16T00:00:00',
            'end_time': '2016-10-24T00:00:00'
        }

        request = self.post_json('/offers', data)

        data['project_id'] = self.context.project_id
        data['uuid'] = self.test_offer.uuid
        data['status'] = statuses.AVAILABLE
        data['parent_lease_uuid'] = None

        mock_gro.assert_called_once_with(self.test_offer.resource_type,
                                         self.test_offer.resource_uuid)
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_create.assert_called_once()
        mock_ogdwai.assert_called_once()
        self.assertEqual(data, request.json)
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.offer.get_resource_object')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.offer.Offer.create')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    def test_post_default_resource_type(self, mock_ogdwai, mock_create,
                                        mock_cra, mock_generate_uuid,
                                        mock_gro):
        resource = IronicNode(self.test_offer_drt.resource_uuid)
        mock_gro.return_value = resource
        mock_generate_uuid.return_value = self.test_offer_drt.uuid
        mock_ogdwai.return_value = self.test_offer_drt.to_dict()

        data = {
            'resource_uuid': self.test_offer_drt.resource_uuid,
            'name': self.test_offer_drt.name,
            'start_time': '2016-07-16T00:00:00',
            'end_time': '2016-10-24T00:00:00'
        }

        request = self.post_json('/offers', data)

        data['project_id'] = self.context.project_id
        data['uuid'] = self.test_offer_drt.uuid
        data['status'] = statuses.AVAILABLE
        data['resource_type'] = 'ironic_node'
        data['parent_lease_uuid'] = None

        mock_gro.assert_called_once_with(self.test_offer_drt.resource_type,
                                         self.test_offer_drt.resource_uuid)
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_create.assert_called_once()
        mock_ogdwai.assert_called_once()
        self.assertEqual(data, request.json)
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.offer.get_resource_object')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.offer.Offer.create')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    def test_post_lessee(self, mock_ogdwai, mock_create, mock_cra,
                         mock_generate_uuid, mock_gpufi, mock_gro):
        resource = TestNode(self.test_offer_lessee.resource_uuid)
        mock_gro.return_value = resource
        mock_gpufi.return_value = 'lessee_uuid'
        mock_generate_uuid.return_value = self.test_offer_lessee.uuid
        mock_create.return_value = self.test_offer_lessee
        mock_ogdwai.return_value = self.test_offer_lessee.to_dict()

        data = {
            'resource_type': self.test_offer_lessee.resource_type,
            'resource_uuid': self.test_offer_lessee.resource_uuid,
            'name': self.test_offer_lessee.name,
            'start_time': '2016-07-16T00:00:00',
            'end_time': '2016-10-24T00:00:00',
            'lessee_id': 'lessee-uuid'
        }

        request = self.post_json('/offers', data)

        data['project_id'] = self.context.project_id
        data['uuid'] = self.test_offer_lessee.uuid
        data['status'] = statuses.AVAILABLE
        data['parent_lease_uuid'] = None

        mock_gro.assert_called_once_with(self.test_offer_lessee.resource_type,
                                         self.test_offer_lessee.resource_uuid)
        mock_gpufi.assert_called_once_with('lessee-uuid')
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_create.assert_called_once()
        mock_ogdwai.assert_called_once()
        self.assertEqual(data, request.json)
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_resource_lease_admin')
    @mock.patch('esi_leap.api.controllers.v1.offer.get_resource_object')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.offer.Offer.create')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    def test_post_non_admin_parent_lease(self, mock_ogdwai, mock_create,
                                         mock_cra, mock_generate_uuid,
                                         mock_gro, mock_crla):
        resource = TestNode(self.test_offer_with_parent.resource_uuid)
        mock_gro.return_value = resource
        mock_generate_uuid.return_value = self.test_offer_with_parent.uuid
        mock_create.return_value = self.test_offer_with_parent
        mock_ogdwai.return_value = self.test_offer_with_parent.to_dict()
        mock_cra.side_effect = exception.HTTPResourceForbidden(
            resource_type='test_node',
            resource=self.test_offer_with_parent.resource_uuid)
        mock_crla.return_value = self.test_offer_with_parent.parent_lease_uuid

        data = {
            'resource_type': self.test_offer_with_parent.resource_type,
            'resource_uuid': self.test_offer_with_parent.resource_uuid,
            'name': self.test_offer_with_parent.name,
            'start_time': '2016-07-16T00:00:00',
            'end_time': '2016-10-24T00:00:00'
        }

        request = self.post_json('/offers', data)

        data['project_id'] = self.context.project_id
        data['uuid'] = self.test_offer_with_parent.uuid
        data['status'] = statuses.AVAILABLE
        data['parent_lease_uuid'] = (
            self.test_offer_with_parent.parent_lease_uuid)

        mock_gro.assert_called_once_with(
            self.test_offer_with_parent.resource_type,
            self.test_offer_with_parent.resource_uuid)
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_crla.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id,
            datetime.datetime(2016, 7, 16, 0, 0, 0),
            datetime.datetime(2016, 10, 24, 0, 0, 0))
        mock_create.assert_called_once()
        mock_ogdwai.assert_called_once()
        self.assertEqual(data, request.json)
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_resource_lease_admin')
    @mock.patch('esi_leap.api.controllers.v1.offer.get_resource_object')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.offer.Offer.create')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    def test_post_non_admin_no_parent_lease(self, mock_ogdwai, mock_create,
                                            mock_cra, mock_generate_uuid,
                                            mock_gro, mock_crla):
        resource = TestNode(self.test_offer_with_parent.resource_uuid)
        mock_gro.return_value = resource
        mock_generate_uuid.return_value = self.test_offer_with_parent.uuid
        mock_create.return_value = self.test_offer_with_parent
        mock_ogdwai.return_value = self.test_offer_with_parent.to_dict()
        mock_cra.side_effect = exception.HTTPResourceForbidden(
            resource_type='test_node',
            resource=self.test_offer_with_parent.resource_uuid)
        mock_crla.return_value = None

        data = {
            'resource_type': self.test_offer_with_parent.resource_type,
            'resource_uuid': self.test_offer_with_parent.resource_uuid,
            'name': self.test_offer_with_parent.name,
            'start_time': '2016-07-16T00:00:00',
            'end_time': '2016-10-24T00:00:00'
        }

        request = self.post_json('/offers', data, expect_errors=True)

        mock_gro.assert_called_once_with(
            self.test_offer_with_parent.resource_type,
            self.test_offer_with_parent.resource_uuid)
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_crla.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id,
            datetime.datetime(2016, 7, 16, 0, 0, 0),
            datetime.datetime(2016, 10, 24, 0, 0, 0))
        mock_create.assert_not_called()
        mock_ogdwai.assert_not_called()
        self.assertEqual(http_client.FORBIDDEN, request.status_int)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    def test_get_nofilters(self, mock_get_all, mock_ogdwai, mock_gpl,
                           mock_gnl):
        mock_get_all.return_value = [self.test_offer, self.test_offer_2]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True)]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        expected_filters = {'status': statuses.OFFER_CAN_DELETE}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}

        request = self.get_json('/offers')

        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 2
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    def test_get_any_status(self, mock_get_all, mock_ogdwai, mock_gpl,
                            mock_gnl):
        mock_get_all.return_value = [self.test_offer, self.test_offer_2]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True)]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        expected_filters = {}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}

        request = self.get_json('/offers/?status=any')

        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 2
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    def test_get_status_filter(self, mock_get_all, mock_ogdwai,
                               mock_gpl, mock_gnl):
        mock_get_all.return_value = [self.test_offer, self.test_offer_2]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True)]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        expected_filters = {'status': [statuses.AVAILABLE]}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}

        request = self.get_json(
            '/offers/?status=available')

        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 2
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    def test_get_project_filter(self, mock_get_all, mock_ogdwai,
                                mock_gpufi, mock_gpl, mock_gnl):
        mock_get_all.return_value = [self.test_offer, self.test_offer_2]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True)]
        mock_gpufi.return_value = self.context.project_id
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        expected_filters = {'project_id': self.context.project_id,
                            'status': statuses.OFFER_CAN_DELETE}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}

        request = self.get_json(
            '/offers/?project_id=' + self.context.project_id)

        mock_gpufi.assert_called_once_with(self.context.project_id)
        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 2
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.offer.get_resource_object')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    def test_get_resource_filter(self, mock_get_all, mock_ogdwai, mock_gro,
                                 mock_gpl, mock_gnl):
        mock_get_all.return_value = [self.test_offer, self.test_offer_2]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True)]
        mock_gro.return_value = TestNode('54321')
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        expected_filters = {'status': statuses.OFFER_CAN_DELETE,
                            'resource_uuid': '54321',
                            'resource_type': 'test_node'}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}

        request = self.get_json('/offers/?resource_uuid=54321&'
                                'resource_type=test_node')

        mock_gro.assert_called_once_with('test_node', '54321')
        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 2
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    def test_get_resource_class_filter(self, mock_get_all, mock_ogdwai,
                                       mock_gpl, mock_gnl):
        mock_get_all.return_value = [self.test_offer,
                                     self.test_offer_2, self.test_offer_drt]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True),
            _get_offer_response(self.test_offer_drt, use_datetime=True)]
        mock_gpl.return_value = []
        mock_gnl.return_value = []
        expected_filters = {'status': statuses.OFFER_CAN_DELETE}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}
        request = self.get_json('/offers/?resource_class=fake')

        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 3
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.offer.get_resource_object')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    def test_get_resource_filter_default_resource_type(self, mock_get_all,
                                                       mock_ogdwai,
                                                       mock_gro,
                                                       mock_gpl,
                                                       mock_gnl):
        fake_uuid = uuidutils.generate_uuid()
        mock_get_all.return_value = [self.test_offer, self.test_offer_2]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True)]
        mock_gro.return_value = IronicNode(fake_uuid)
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        expected_filters = {'status': statuses.OFFER_CAN_DELETE,
                            'resource_uuid': fake_uuid,
                            'resource_type': 'ironic_node'}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}

        request = self.get_json('/offers/?resource_uuid=%s' % fake_uuid)
        mock_gro.assert_called_once_with('ironic_node', fake_uuid)
        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 2
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.offer.Offer.get_all')
    @mock.patch('esi_leap.api.controllers.v1.utils.policy_authorize')
    def test_get_lessee_filter(self, mock_authorize, mock_get_all,
                               mock_ogdwai, mock_gpl, mock_gnl):
        mock_get_all.return_value = [self.test_offer, self.test_offer_2]
        mock_ogdwai.side_effect = [
            _get_offer_response(self.test_offer, use_datetime=True),
            _get_offer_response(self.test_offer_2, use_datetime=True)]
        mock_authorize.side_effect = [
            None,
            exception.HTTPForbidden(rule='esi_leap:offer:get')
        ]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        expected_filters = {'status': statuses.OFFER_CAN_DELETE,
                            'lessee_id': self.context.project_id}
        expected_resp = {'offers': [_get_offer_response(self.test_offer),
                                    _get_offer_response(self.test_offer_2)]}

        request = self.get_json('/offers')

        mock_get_all.assert_called_once_with(expected_filters, self.context)
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        assert mock_ogdwai.call_count == 2
        self.assertEqual(request, expected_resp)

    @mock.patch('esi_leap.api.controllers.v1.utils.check_offer_lessee')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_offer_policy_and_retrieve')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'offer_get_dict_with_added_info')
    def test_get_one(self, mock_ogdwai, mock_copar, mock_col):
        mock_copar.return_value = self.test_offer
        mock_ogdwai.return_value = self.test_offer.to_dict()

        self.get_json('/offers/' + self.test_offer.uuid)

        mock_copar.assert_called_once_with(self.context,
                                           'esi_leap:offer:get',
                                           self.test_offer.uuid)
        mock_col.assert_called_once_with(self.context.to_policy_values(),
                                         self.test_offer)
        mock_ogdwai.assert_called_once_with(self.test_offer)

    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.objects.lease.Lease.create')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_offer_lessee')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_offer_policy_and_retrieve')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    def test_claim(self, mock_lgdwai, mock_copar, mock_col, mock_lease_create,
                   mock_generate_uuid):
        lease_uuid = '12345'
        mock_generate_uuid.return_value = lease_uuid
        mock_copar.return_value = self.test_offer
        data = {
            'name': 'lease_claim',
            'start_time': '2016-07-16T19:20:30',
            'end_time': '2016-08-16T19:20:30'
        }

        request = self.post_json('/offers/' + self.test_offer.uuid + '/claim',
                                 data)

        mock_copar.assert_called_once_with(self.context,
                                           'esi_leap:offer:claim',
                                           self.test_offer.uuid,
                                           [statuses.AVAILABLE])
        mock_col.assert_called_once_with(self.context.to_policy_values(),
                                         self.test_offer)
        mock_lease_create.assert_called_once()
        mock_lgdwai.assert_called_once()
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.objects.lease.Lease.create')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_offer_lessee')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_offer_policy_and_retrieve')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    def test_claim_parent_lease(self, mock_lgdwai, mock_copar, mock_col,
                                mock_lease_create, mock_generate_uuid):
        lease_uuid = '12345'
        mock_generate_uuid.return_value = lease_uuid
        mock_copar.return_value = self.test_offer_with_parent
        data = {
            'name': 'lease_claim',
            'start_time': '2016-07-16T19:20:30',
            'end_time': '2016-08-16T19:20:30'
        }

        request = self.post_json(
            '/offers/' + self.test_offer_with_parent.uuid + '/claim',
            data)

        mock_copar.assert_called_once_with(self.context,
                                           'esi_leap:offer:claim',
                                           self.test_offer_with_parent.uuid,
                                           [statuses.AVAILABLE])
        mock_col.assert_called_once_with(self.context.to_policy_values(),
                                         self.test_offer_with_parent)
        mock_lease_create.assert_called_once()
        mock_lgdwai.assert_called_once()
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_offer_policy_and_retrieve')
    @mock.patch('esi_leap.objects.offer.Offer.cancel')
    def test_delete(self, mock_cancel, mock_copar):
        mock_copar.return_value = self.test_offer

        self.delete_json('/offers/' + self.test_offer.uuid)

        mock_copar.assert_called_once_with(self.context,
                                           'esi_leap:offer:delete',
                                           self.test_offer.uuid,
                                           statuses.OFFER_CAN_DELETE)
        mock_cancel.assert_called_once()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from esi_leap.common.i18n import _
from oslo_config import cfg

# borrowed from Ironic
opts = [
    cfg.StrOpt('notification_level',
               choices=[('debug', _('"debug" level')),
                        ('info', _('"info" level')),
                        ('warning', _('"warning" level')),
                        ('error', _('"error" level')),
                        ('critical', _('"critical" level'))],
               help=_('Specifies the minimum level for which to send '
                      'notifications. If not set, no notifications will '
                      'be sent.')),
    cfg.ListOpt(
        'versioned_notifications_topics',
        default=['esi_leap_versioned_notifications'],
        help=_('Specifies the topics for '
               'the versioned notifications issued by esi-leap.')),
]


notification_group = cfg.OptGroup('notification', title='Notification Options')


def register_opts(conf):
    conf.register_opts(opts, group=notification_group)
    conf.set_default('notification_level', 'info', group=notification_group)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from esi_leap.common import exception
from esi_leap.common import notification_utils as notify
from esi_leap.common import statuses
from esi_leap.common import utils
from esi_leap.db import api as dbapi
from esi_leap.objects import base
from esi_leap.objects import fields
from esi_leap.objects import notification
from esi_leap.objects import offer as offer_obj
from esi_leap.resource_objects import get_resource_object

from oslo_config import cfg
from oslo_log import log as logging
from oslo_versionedobjects import base as versioned_objects_base

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


@versioned_objects_base.VersionedObjectRegistry.register
class LeaseCRUDNotification(notification.NotificationBase):
    """Notification emitted when a lease is created or deleted."""

    fields = {
        'payload': fields.ObjectField('LeaseCRUDPayload')
    }


@versioned_objects_base.VersionedObjectRegistry.register
class LeaseCRUDPayload(notification.NotificationPayloadBase):
    """Payload schema for when a lease is created or deleted."""
    # Version 1.0: Initial version
    VERSION = '1.0'

    SCHEMA = {
        'id': ('lease', 'id'),
        'name': ('lease', 'name'),
        'uuid': ('lease', 'uuid'),
        'project_id': ('lease', 'project_id'),
        'owner_id': ('lease', 'owner_id'),
        'resource_type': ('lease', 'resource_type'),
        'resource_uuid': ('lease', 'resource_uuid'),
        'start_time': ('lease', 'start_time'),
        'end_time': ('lease', 'end_time'),
        'fulfill_time': ('lease', 'fulfill_time'),
        'expire_time': ('lease', 'expire_time'),
        'status': ('lease', 'status'),
        'properties': ('lease', 'properties'),
        'purpose': ('lease', 'purpose'),
        'offer_uuid': ('lease', 'offer_uuid'),
        'parent_lease_uuid': ('lease', 'parent_lease_uuid'),
        'node_name': ('node', 'node_name'),
        'node_uuid': ('node', '_uuid'),
        'node_provision_state': ('node', 'node_provision_state'),
        'node_power_state': ('node', 'node_power_state'),
        'node_properties': ('node', 'node_properties'),
    }

    fields = {
        'id': fields.IntegerField(),
        'name': fields.StringField(nullable=True),
        'uuid': fields.UUIDField(),
        'project_id': fields.StringField(),
        'owner_id': fields.StringField(),
        'purpose': fields.StringField(nullable=True),
        'resource_type': fields.StringField(),
        'resource_uuid': fields.StringField(),
        'start_time': fields.DateTimeField(nullable=True),
        'end_time': fields.DateTimeField(nullable=True),
        'fulfill_time': fields.DateTimeField(nullable=True),
        'expire_time': fields.DateTimeField(nullable=True),
        'status': fields.StringField(),
        'properties': fields.FlexibleDictField(nullable=True),
        'offer_uuid': fields.UUIDField(nullable=True),
        'parent_lease_uuid': fields.UUIDField(nullable=True),
        'node_name': fields.StringField(nullable=True),
        'node_uuid': fields.UUIDField(),
        'node_provision_state': fields.StringField(),
        'node_power_state': fields.StringField(),
        'node_properties': fields.FlexibleDictField(nullable=True),
    }

    def __init__(self, lease, node):
        super(LeaseCRUDPayload, self).__init__()

        setattr(node, 'node_name', node.get_name())
        setattr(node, 'node_provision_state', node.get_node_provision_state())
        setattr(node, 'node_power_state', node.get_node_power_state())
        node_config = node.get_config().copy()
        node_config.pop('lease_uuid', None)
        setattr(node, 'node_properties', node_config)

        self.populate_schema(lease=lease, node=node)

    def get_event_dict(self, event_type):
        event_dict = super().get_event_dict(event_type)
        event_dict.update({
            'object_type': 'lease',
            'object_uuid': self.uuid,
            'resource_type': self.resource_type,
            'resource_uuid': self.resource_uuid,
            'lessee_id': self.project_id,
            'owner_id': self.owner_id,
        })
        return event_dict


CRUD_NOTIFY_OBJ = {
    'lease': (LeaseCRUDNotification, LeaseCRUDPayload),
}


@versioned_objects_base.VersionedObjectRegistry.register
class Lease(base.ESILEAPObject):
    dbapi = dbapi.get_instance()

    fields = {
        'id': fields.IntegerField(),
        'name': fields.StringField(nullable=True),
        'uuid': fields.UUIDField(),
        'project_id': fields.StringField(),
        'owner_id': fields.StringField(),
        'purpose': fields.StringField(nullable=True),
        'resource_type': fields.StringField(),
        'resource_uuid': fields.StringField(),
        'start_time': fields.DateTimeField(nullable=True),
        'end_time': fields.DateTimeField(nullable=True),
        'fulfill_time': fields.DateTimeField(nullable=True),
        'expire_time': fields.DateTimeField(nullable=True),
        'status': fields.StringField(),
        'properties': fields.FlexibleDictField(nullable=True),
        'offer_uuid': fields.UUIDField(nullable=True),
        'parent_lease_uuid': fields.UUIDField(nullable=True),
    }

    @classmethod
    def get(cls, lease_uuid, context=None):
        db_lease = cls.dbapi.lease_get_by_uuid(lease_uuid)
        if db_lease:
            return cls._from_db_object(context, cls(), db_lease)

    @classmethod
    def get_all(cls, filters, context=None):
        db_leases = cls.dbapi.lease_get_all(filters)
        return cls._from_db_object_list(context, db_leases)

    def create(self, context=None):
        updates = self.obj_get_changes()

        with utils.lock(utils.get_resource_lock_name(updates['resource_type'],
                                                     updates['resource_uuid']),
                        external=True):
            if updates['start_time'] >= updates['end_time']:
                raise exception.InvalidTimeRange(
                    resource='lease',
                    start_time=str(updates['start_time']),
                    end_time=str(updates['end_time'])
                    )

            # check availability
            if 'offer_uuid' in updates:
                # lease is being created from an offer
                related_offer = offer_obj.Offer.get(updates['offer_uuid'])

                if related_offer.status != statuses.AVAILABLE:
                    raise exception.OfferNotAvailable(
                        offer_uuid=related_offer.uuid,
                        status=related_offer.status)

                related_offer.verify_availability(updates['start_time'],
                                                  updates['end_time'])
            elif 'parent_lease_uuid' in updates:
                # lease is a child of an existing lease
                parent_lease = Lease.get(updates['parent_lease_uuid'])

                if parent_lease.status != statuses.ACTIVE:
                    raise exception.LeaseNotActive(
                        updates['parent_lease_uuid'])

                parent_lease.verify_child_availability(updates['start_time'],
                                                       updates['end_time'])
            else:
                ro = get_resource_object(updates['resource_type'],
                                         updates['resource_uuid'])
                ro.verify_availability(updates['start_time'],
                                       updates['end_time'])

            db_lease = self.dbapi.lease_create(updates)
            self._from_db_object(context, self, db_lease)

    def cancel(self, context=None):
        leases = Lease.get_all(
            {'parent_lease_uuid': self.uuid,
             'status': statuses.LEASE_CAN_DELETE},
            None)
        for lease in leases:
            lease.cancel()
        offers = offer_obj.Offer.get_all(
            {'parent_lease_uuid': self.uuid,
             'status': statuses.OFFER_CAN_DELETE},
            None)
        for offer in offers:
            offer.cancel()

        with utils.lock(utils.get_resource_lock_name(self.resource_type,
                                                     self.resource_uuid),
                        external=True):
            LOG.info('Deleting lease %s', self.uuid)
            try:
                resource = self.resource_object()
                self.deactivate(context, resource)
                self.status = statuses.DELETED
                self.expire_time = datetime.datetime.now()

            except Exception as e:
                LOG.info('Error canceling lease: %s: %s' %
                         (type(e).__name__, e))
                LOG.info('Setting lease status to WAIT')
                self.status = statuses.WAIT_CANCEL
            self.save(context)

    def destroy(self):
        self.dbapi.lease_destroy(self.uuid)
        self.obj_reset_changes()

    def save(self, context=None):
        updates = self.obj_get_changes()
        db_lease = self.dbapi.lease_update(
            self.uuid, updates)
        self._from_db_object(context, self, db_lease)

    def fulfill(self, context=None):
        with utils.lock(utils.get_resource_lock_name(self.resource_type,
                                                     self.resource_uuid),
                        external=True):
            LOG.info('Fulfilling lease %s', self.uuid)
            try:
                resource = self.resource_object()
                notify.emit_start_notification(context, self,
                                               'fulfill',
                                               CRUD_NOTIFY_OBJ,
                                               node=resource)
                with notify.handle_error_notification(context,
                                                      self,
                                                      'fulfill',
                                                      CRUD_NOTIFY_OBJ,
                                                      node=resource):
                    resource.set_lease(self)

                notify.emit_end_notification(context, self,
                                             'fulfill',
                                             CRUD_NOTIFY_OBJ,
                                             node=resource)

                self.status = statuses.ACTIVE
                self.fulfill_time = datetime.datetime.now()

            except Exception as e:
                LOG.info('Error fulfilling lease: %s: %s' %
                         (type(e).__name__, e))
                LOG.info('Setting lease status to WAIT')
                self.status = statuses.WAIT_FULFILL
            self.save(context)

    def expire(self, context=None):
        leases = Lease.get_all(
            {'parent_lease_uuid': self.uuid,
             'status': statuses.LEASE_CAN_DELETE},
            None)
        for lease in leases:
            lease.expire(context)
        offers = offer_obj.Offer.get_all(
            {'parent_lease_uuid': self.uuid,
             'status': statuses.OFFER_CAN_DELETE},
            None)
        for offer in offers:
            offer.expire(context)

        with utils.lock(utils.get_resource_lock_name(self.resource_type,
                                                     self.resource_uuid),
                        external=True):
            LOG.info('Expiring lease %s', self.uuid)
            try:
                # expire lease
                resource = self.resource_object()
                self.deactivate(context, resource)
                self.status = statuses.EXPIRED
                self.expire_time = datetime.datetime.now()

            except Exception as e:
                LOG.info('Error expiring lease: %s: %s' %
                         (type(e).__name__, e))
                LOG.info('Setting lease status to WAIT')
                self.status = statuses.WAIT_EXPIRE
            self.save(context)

    def resource_object(self):
        return get_resource_object(self.resource_type, self.resource_uuid)

    def verify_child_availability(self, start_time, end_time):
        return self.dbapi.lease_verify_child_availability(
            self, start_time, end_time)

    def deactivate(self, context, resource):
        notify.emit_start_notification(context, self,
                                       'delete', CRUD_NOTIFY_OBJ,
                                       node=resource)

        with notify.handle_error_notification(context,
                                              self,
                                              'delete',
                                              CRUD_NOTIFY_OBJ,
                                              node=resource):
            if resource.get_lease_uuid() == self.uuid:
                resource.remove_lease(self)

                if self.parent_lease_uuid is not None:
                    parent_lease = Lease.get(self.parent_lease_uuid)
                    resource.set_lease(parent_lease)

        notify.emit_end_notification(context, self,
                                     'delete', CRUD_NOTIFY_OBJ,
                                     node=resource)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from esi_leap.common import exception
from esi_leap.common import statuses
from esi_leap.common import utils
from esi_leap.db import api as dbapi
from esi_leap.objects import base
from esi_leap.objects import fields
from esi_leap.objects import lease as lease_obj
from esi_leap.resource_objects import get_resource_object

from oslo_config import cfg
from oslo_log import log as logging
from oslo_versionedobjects import base as versioned_objects_base

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


@versioned_objects_base.VersionedObjectRegistry.register
class Offer(base.ESILEAPObject):
    dbapi = dbapi.get_instance()

    fields = {
        'id': fields.IntegerField(),
        'name': fields.StringField(nullable=True),
        'uuid': fields.UUIDField(),
        'project_id': fields.StringField(),
        'lessee_id': fields.StringField(nullable=True),
        'resource_type': fields.StringField(),
        'resource_uuid': fields.StringField(),
        'start_time': fields.DateTimeField(nullable=True),
        'end_time': fields.DateTimeField(nullable=True),
        'status': fields.StringField(),
        'properties': fields.FlexibleDictField(nullable=True),
        'parent_lease_uuid': fields.UUIDField(nullable=True),
    }

    @classmethod
    def get(cls, offer_uuid, context=None):
        db_offer = cls.dbapi.offer_get_by_uuid(offer_uuid)
        if db_offer:
            return cls._from_db_object(context, cls(), db_offer)

    @classmethod
    def get_all(cls, filters, context=None):
        db_offers = cls.dbapi.offer_get_all(filters)
        return cls._from_db_object_list(context, db_offers)

    def get_availabilities(self):

        if self.status != statuses.AVAILABLE:
            return []

        conflicts = self.dbapi.offer_get_conflict_times(self)
        now = datetime.datetime.now()
        start_time = self.start_time if self.start_time >= now else now

        if conflicts:
            # if the conflicts are all in the past:
            if conflicts[-1][1] <= start_time:
                avails = [[start_time, self.end_time]]
            else:
                # if the offer starts in the future or
                # the first conflict time is in the future,
                # add the $start_time and first conflict start_time
                # in the array
                if start_time < conflicts[0][0]:
                    times = [start_time, conflicts[0][0]]
                    for i in range(len(conflicts) - 1):
                        times.append(conflicts[i][1])
                        times.append(conflicts[i + 1][0])
                else:
                    times = []
                    for i in range(len(conflicts) - 1):
                        # Find the conflict timeframe that started
                        # in the past and will end in the future.
                        # add all times after this
                        if (conflicts[i][0] <= start_time and
                            conflicts[i][1] > start_time) \
                            or conflicts[i][0] > start_time:
                            times.append(conflicts[i][1])
                            times.append(conflicts[i + 1][0])

                times.append(conflicts[-1][1])
                times.append(self.end_time)

                i = 0
                while i < len(times) - 1:
                    if times[i] == times[i + 1]:
                        times.pop(i)
                        times.pop(i)
                    else:
                        i += 1

                avails = [[times[j], times[j + 1]]
                          for j in range(0, len(times) - 1, 2)]

        else:
            avails = [[start_time, self.end_time]]

        return avails

    def get_next_lease_start_time(self, start):
        return self.dbapi.offer_get_next_lease_start_time(
            self.uuid, start)

    def create(self, context=None):
        updates = self.obj_get_changes()

        with utils.lock(utils.get_resource_lock_name(updates['resource_type'],
                                                     updates['resource_uuid']),
                        external=True):
            LOG.info('Creating offer')
            if updates['start_time'] >= updates['end_time']:
                raise exception.InvalidTimeRange(
                    resource='offer',
                    start_time=str(updates['start_time']),
                    end_time=str(updates['end_time'])
                )

            if updates.get('parent_lease_uuid'):
                # offer is a child of an existing lease
                parent_lease = lease_obj.Lease.get(
                    updates['parent_lease_uuid'])

                if parent_lease.status != statuses.ACTIVE:
                    raise exception.LeaseNotActive(
                        updates['parent_lease_uuid'])

                parent_lease.verify_child_availability(updates['start_time'],
                                                       updates['end_time'])
            else:
                ro = get_resource_object(updates['resource_type'],
                                         updates['resource_uuid'])
                ro.verify_availability(updates['start_time'],
                                       updates['end_time'])

            db_offer = self.dbapi.offer_create(updates)
            self._from_db_object(context, self, db_offer)

    def cancel(self):
        LOG.info('Deleting offer %s', self.uuid)
        leases = lease_obj.Lease.get_all(
            {'offer_uuid': self.uuid,
             'status': statuses.LEASE_CAN_DELETE},
            None)
        for lease in leases:
            lease.cancel()

        with utils.lock(utils.get_resource_lock_name(self.resource_type,
                                                     self.resource_uuid),
                        external=True):
            self.status = statuses.DELETED
            self.save(None)

    def expire(self, context=None):
        LOG.info('Expiring offer %s', self.uuid)
        leases = lease_obj.Lease.get_all(
            {'offer_uuid': self.uuid,
             'status': statuses.LEASE_CAN_DELETE},
            None)
        for lease in leases:
            lease.expire(context)

        with utils.lock(utils.get_resource_lock_name(self.resource_type,
                                                     self.resource_uuid),
                        external=True):
            self.status = statuses.EXPIRED
            self.save(context)

    def verify_availability(self, start_time, end_time):
        return self.dbapi.offer_verify_availability(
            self, start_time, end_time)

    def destroy(self):
        self.dbapi.offer_destroy(self.uuid)
        self.obj_reset_changes()

    def save(self, context=None):
        updates = self.obj_get_changes()
        db_offer = self.dbapi.offer_update(
            self.uuid, updates)
        self._from_db_object(context, self, db_offer)

    def resource_object(self):
        return get_resource_object(self.resource_type, self.resource_uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironicclient.common.apiclient import exceptions as ir_exception
from oslo_log import log as logging
from oslo_utils.uuidutils import is_uuid_like

from esi_leap.common import exception
from esi_leap.common import ironic
import esi_leap.conf
from esi_leap.resource_objects import base
from esi_leap.resource_objects import error


CONF = esi_leap.conf.CONF
_cached_ironic_client = None

LOG = logging.getLogger(__name__)


def get_ironic_client():
    global _cached_ironic_client
    if _cached_ironic_client is None:
        _cached_ironic_client = ironic.get_ironic_client()
    return _cached_ironic_client


class IronicNode(base.ResourceObjectInterface):

    resource_type = 'ironic_node'

    def __init__(self, ident):
        if not is_uuid_like(ident):
            self._node = get_ironic_client().node.get(ident)
            self._uuid = self._node.uuid
        else:
            self._node = None
            self._uuid = ident

    def get_uuid(self):
        return self._uuid

    def get_name(self, resource_list=None):
        return self._get_node_attr('name', '',
                                   resource_list=resource_list,
                                   err_msg='Error getting resource name',
                                   err_val=error.UNKNOWN['name'])

    def get_resource_class(self, resource_list=None):
        return self._get_node_attr('resource_class', '',
                                   resource_list=resource_list,
                                   err_msg='Error getting resource class',
                                   err_val=error.UNKNOWN['resource_class'])

    def get_config(self):
        config = self._get_node_attr('properties', {},
                                     err_msg='Error getting resource config',
                                     err_val=error.UNKNOWN['config'])
        return config

    def get_owner_project_id(self):
        return self._get_node_attr('owner', '',
                                   err_msg='Error getting owner project id',
                                   err_val=error.UNKNOWN['owner_project_id'])

    def get_lease_uuid(self):
        props = self._get_node_attr('properties', None,
                                    err_msg='Error getting lease UUID',
                                    err_val=error.UNKNOWN['lease_uuid'])
        return None if props is None else props.get('lease_uuid', None)

    def get_lessee_project_id(self):
        return self._get_node_attr('lessee', '',
                                   err_msg='Error getting lessee project id',
                                   err_val=error.UNKNOWN['lessee_project_id'])

    def get_node_provision_state(self):
        return self._get_node_attr('provision_state', '',
                                   err_msg='Error getting provision state',
                                   err_val=error.UNKNOWN['provision_state'])

    def get_node_power_state(self):
        return self._get_node_attr('power_state', '',
                                   err_msg='Error getting power state',
                                   err_val=error.UNKNOWN['power_state'])

    def set_lease(self, lease):
        patches = []
        patches.append({
            'op': 'add',
            'path': '/properties/lease_uuid',
            'value': lease.uuid,
        })
        patches.append({
            'op': 'add',
            'path': '/lessee',
            'value': lease.project_id,
        })
        get_ironic_client().node.update(self._uuid, patches)

    def remove_lease(self, lease):
        patches = []
        uuid = self.get_lease_uuid()
        if uuid != lease.uuid:
            return
        if uuid:
            patches.append({
                'op': 'remove',
                'path': '/properties/lease_uuid',
            })
        if self.get_lessee_project_id():
            patches.append({
                'op': 'remove',
                'path': '/lessee',
            })
        if len(patches) > 0:
            get_ironic_client().node.update(self._uuid, patches)
        state = self._get_node().provision_state
        if state == 'active':
            get_ironic_client().node.set_provision_state(self._uuid, 'deleted')

    def _get_node(self, resource_list=None):
        try:
            if not self._node:
                self._node = ironic.get_node(self._uuid, resource_list)
        except ir_exception.NotFound as e:
            raise exception.NodeNotFound(uuid=self._uuid,
                                         resource_type=self.resource_type,
                                         err=str(e))
        return self._node

    def _get_node_attr(self, attr, default=None, resource_list=None,
                       err_val=None, err_msg=None):
        try:
            return getattr(self._get_node(resource_list), attr, default)
        except exception.NodeNotFound:
            LOG.exception(err_msg)
            return err_val if err_val is not None else default



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy as oslo_policy
from oslo_utils import uuidutils

import datetime

from esi_leap.common import exception
from esi_leap.common import keystone
from esi_leap.common import policy
from esi_leap.objects import lease as lease_obj
from esi_leap.objects import offer as offer_obj


def check_resource_admin(cdict, resource, project_id):
    if project_id != resource.get_owner_project_id():
        resource_policy_authorize('esi_leap:offer:offer_admin',
                                  cdict, cdict,
                                  resource.resource_type,
                                  resource.get_uuid())


def check_resource_lease_admin(cdict, resource, project_id,
                               start_time, end_time):
    # check to see if project is current lessee
    if project_id == resource.get_lessee_project_id():
        parent_lease_uuid = resource.get_lease_uuid()
        if parent_lease_uuid is not None:
            parent_lease = lease_obj.Lease.get(parent_lease_uuid)

            # don't allow sub-sub-leases
            if parent_lease.parent_lease_uuid is None:
                # check if offer is within start and end time bounds
                if ((start_time >= parent_lease.start_time) and
                        (end_time <= parent_lease.end_time)):
                    return parent_lease_uuid
                else:
                    raise exception.ResourceNoPermissionTime(
                        resource_type=resource.resource_type,
                        resource_uuid=resource.get_uuid(),
                        start_time=start_time,
                        end_time=end_time)


def get_offer(uuid_or_name, status_filters=[]):
    if uuidutils.is_uuid_like(uuid_or_name):
        o = offer_obj.Offer.get(uuid_or_name)
        if not status_filters or o.status in status_filters:
            return o
        else:
            raise exception.OfferNotFound(offer_uuid=uuid_or_name)
    else:
        offers = offer_obj.Offer.get_all({'name': uuid_or_name,
                                          'status': status_filters})

        if len(offers) != 1:
            if len(offers) == 0:
                raise exception.OfferNotFound(offer_uuid=uuid_or_name)
            else:
                raise exception.OfferDuplicateName(name=uuid_or_name)

        return offers[0]


def get_lease(uuid_or_name, status_filters=[]):
    if uuidutils.is_uuid_like(uuid_or_name):
        lease = lease_obj.Lease.get(uuid_or_name)
        if not status_filters or lease.status in status_filters:
            return lease
        else:
            raise exception.LeaseNotFound(lease_id=uuid_or_name)
    else:
        leases = lease_obj.Lease.get_all({'name': uuid_or_name,
                                          'status': status_filters})

        if len(leases) != 1:
            if len(leases) == 0:
                raise exception.LeaseNotFound(lease_id=uuid_or_name)
            else:
                raise exception.LeaseDuplicateName(name=uuid_or_name)

        return leases[0]


def policy_authorize(policy_name, target, creds):
    try:
        policy.authorize(policy_name, target, creds)
    except oslo_policy.PolicyNotAuthorized:
        raise exception.HTTPForbidden(rule=policy_name)


def resource_policy_authorize(policy_name, target, creds,
                              resource_type, resource):
    try:
        policy_authorize(policy_name, target, creds)
    except exception.HTTPForbidden:
        raise exception.HTTPResourceForbidden(resource_type=resource_type,
                                              resource=resource)


def check_lease_policy_and_retrieve(request, policy_name, lease_ident,
                                    status_filters=[]):
    lease = get_lease(lease_ident, status_filters)

    cdict = request.to_policy_values()
    target = dict(cdict)
    target['lease.owner_id'] = lease.owner_id
    target['lease.project_id'] = lease.project_id

    resource_policy_authorize(policy_name, target, cdict, 'lease', lease.uuid)
    return lease


def check_offer_policy_and_retrieve(request, policy_name, offer_ident,
                                    status_filters=[]):
    offer = get_offer(offer_ident, status_filters)

    cdict = request.to_policy_values()
    target = dict(cdict)
    target['offer.project_id'] = offer.project_id

    resource_policy_authorize(policy_name, target, cdict, 'offer', offer.uuid)
    return offer


def check_offer_lessee(cdict, offer):
    project_id = cdict['project_id']

    # pass if offer has no lessee limitation or project_id created the offer
    if offer.lessee_id is None or offer.project_id == project_id:
        return

    if offer.lessee_id not in keystone.get_parent_project_id_tree(project_id):
        resource_policy_authorize(
            'esi_leap:offer:offer_admin',
            cdict, cdict, 'offer', offer.uuid)


def offer_get_dict_with_added_info(offer, project_list=None, node_list=None):
    resource = offer.resource_object()

    o = offer.to_dict()
    o['availabilities'] = offer.get_availabilities()
    o['project'] = keystone.get_project_name(offer.project_id, project_list)
    o['lessee'] = keystone.get_project_name(offer.lessee_id, project_list)
    o['resource'] = resource.get_name(node_list)
    o['resource_class'] = resource.get_resource_class(node_list)
    return o


def lease_get_dict_with_added_info(lease, project_list=None, node_list=None):
    resource = lease.resource_object()

    lease_dict = lease.to_dict()
    lease_dict['project'] = keystone.get_project_name(lease.project_id,
                                                      project_list)
    lease_dict['owner'] = keystone.get_project_name(lease.owner_id,
                                                    project_list)
    lease_dict['resource'] = resource.get_name(node_list)
    lease_dict['resource_class'] = resource.get_resource_class(node_list)
    return lease_dict


def check_lease_length(cdict, start_time, end_time, max_time):
    if (end_time - start_time) > datetime.timedelta(days=max_time):
        # Check if the current project is admin
        # Only admin project can set a lease beyond the max length
        try:
            policy_authorize('esi_leap:lease:lease_admin', cdict, cdict)
        except exception.HTTPForbidden:
            raise exception.LeaseExceedMaxTimeRange(max_time=max_time)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import wsme
from wsme import types as wtypes


class ESILEAPBase(wtypes.Base):

    def to_dict(self):
        esi_leap_dict = {}
        if self.fields:
            for key in self.fields:
                val = getattr(self, key, wsme.Unset)
                if val != wsme.Unset:
                    esi_leap_dict[key] = val
        return esi_leap_dict



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from esi_leap.objects import fields
from esi_leap.tests import base


# FlexibleDict borrowed from Ironic
class TestFlexibleDictField(base.TestCase):

    def setUp(self):
        super(TestFlexibleDictField, self).setUp()
        self.field = fields.FlexibleDictField()

    def test_coerce(self):
        d = {'foo_1': 'bar', 'foo_2': 2, 'foo_3': [], 'foo_4': {}}
        self.assertEqual(d, self.field.coerce('obj', 'attr', d))
        self.assertEqual({'foo': 'bar'},
                         self.field.coerce('obj', 'attr', '{"foo": "bar"}'))

    def test_coerce_bad_values(self):
        self.assertRaises(TypeError, self.field.coerce, 'obj', 'attr', 123)
        self.assertRaises(TypeError, self.field.coerce, 'obj', 'attr', True)

    def test_coerce_nullable_translation(self):
        # non-nullable
        self.assertRaises(ValueError, self.field.coerce, 'obj', 'attr', None)

        # nullable
        self.field = fields.FlexibleDictField(nullable=True)
        self.assertEqual({}, self.field.coerce('obj', 'attr', None))


# NotificationLevelField borrowed from Ironic
class TestNotificationLevelField(base.TestCase):

    def setUp(self):
        super(TestNotificationLevelField, self).setUp()
        self.field = fields.NotificationLevelField()

    def test_coerce_good_value(self):
        self.assertEqual(fields.NotificationLevel.WARNING,
                         self.field.coerce('obj', 'attr', 'warning'))

    def test_coerce_bad_value(self):
        self.assertRaises(ValueError, self.field.coerce, 'obj', 'attr',
                          'not_a_priority')


# NotificationStatusField borrowed from Ironic
class TestNotificationStatusField(base.TestCase):

    def setUp(self):
        super(TestNotificationStatusField, self).setUp()
        self.field = fields.NotificationStatusField()

    def test_coerce_good_value(self):
        self.assertEqual(fields.NotificationStatus.START,
                         self.field.coerce('obj', 'attr', 'start'))

    def test_coerce_bad_value(self):
        self.assertRaises(ValueError, self.field.coerce, 'obj', 'attr',
                          'not_a_priority')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Borrowed from Ironic

import json
from wsme import types as wtypes


class JsonType(wtypes.UserType):
    """A simple JSON type."""

    basetype = wtypes.text
    name = 'json'

    def __str__(self):
        # These are the json serializable native types
        return ' | '.join(map(str, (wtypes.text, int, float,
                                    bool, list, dict, None)))

    @staticmethod
    def validate(value):
        json.dumps(value)
        return value

    @staticmethod
    def frombasetype(value):
        return JsonType.validate(value)


class Collection(wtypes.Base):

    @property
    def collection(self):
        return getattr(self, self._type)

    def has_next(self, limit):
        """Return whether collection has more items."""
        return len(self.collection) and len(self.collection) == limit

    def get_next(self, limit, url=None, **kwargs):
        if not self.has_next(limit):
            return wtypes.Unset

        url = url or self._type
        q_args = ''.join(['%s=%s&' % item for item in kwargs.items()])
        next_args = '?%(args)slimit=%(limit)d&marker=%(marker)s' % {
            'args': q_args, 'limit': limit,
            'marker': getattr(self.collection[-1], 'uuid')}

        next_link = '%(url)s/v1/%(resource)s%(args)s' % {
            'url': url,
            'resource': self._type,
            'args': next_args
        }

        return next_link


jsontype = JsonType()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from keystoneauth1 import loading
from oslo_config import cfg


opts = []
ironic_group = cfg.OptGroup('ironic', title='Ironic Options')


def register_opts(conf):
    conf.register_opts(opts, group=ironic_group)
    loading.register_session_conf_options(conf, ironic_group)
    loading.register_auth_conf_options(conf, ironic_group)
    loading.register_adapter_conf_options(conf, ironic_group)
    conf.set_default('valid_interfaces', ['internal', 'public'],
                     group=ironic_group)
    conf.set_default('service_type', 'baremetal', group=ironic_group)


def list_opts():
    def add_options(opts, opts_to_add):
        for new_opt in opts_to_add:
            for opt in opts:
                if opt.name == new_opt.name:
                    break
            else:
                opts.append(new_opt)

    opts_copy = copy.deepcopy(opts)
    opts_copy.insert(0, loading.get_auth_common_conf_options()[0])

    plugins = ['password', 'v2password', 'v3password']
    for name in plugins:
        plugin = loading.get_plugin_loader(name)
        add_options(opts_copy, loading.get_auth_plugin_conf_options(plugin))
    add_options(opts_copy, loading.get_session_conf_options())

    adapter_opts = loading.get_adapter_conf_options(
        include_deprecated=False)
    # adding defaults for valid interfaces
    cfg.set_defaults(adapter_opts, service_type='baremetal',
                     valid_interfaces=['internal', 'public'])
    add_options(opts_copy, adapter_opts)
    opts_copy.sort(key=lambda x: x.name)
    return opts_copy



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_log import log as logging

from esi_leap.api.app import WSGIApplication
from esi_leap.common import i18n
from esi_leap.common import service
import esi_leap.conf


CONF = esi_leap.conf.CONF
LOG = logging.getLogger(__name__)


def initialize_wsgi_app(argv=sys.argv):
    i18n.install('esi_leap')

    service.prepare_service(argv)

    LOG.debug('Configuration:')
    CONF.log_opt_values(LOG, logging.DEBUG)

    return WSGIApplication()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import concurrent.futures
import datetime
import http.client as http_client
from oslo_utils import uuidutils
import pecan
from pecan import rest
import wsme
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from esi_leap.api.controllers import base
from esi_leap.api.controllers import types
from esi_leap.api.controllers.v1 import lease
from esi_leap.api.controllers.v1 import utils
from esi_leap.common import exception
from esi_leap.common import ironic
from esi_leap.common import keystone
from esi_leap.common import statuses
import esi_leap.conf
from esi_leap.objects import lease as lease_obj
from esi_leap.objects import offer as offer_obj
from esi_leap.resource_objects import get_resource_object

CONF = esi_leap.conf.CONF


class Offer(base.ESILEAPBase):

    name = wsme.wsattr(wtypes.text)
    uuid = wsme.wsattr(wtypes.text, readonly=True)
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    project = wsme.wsattr(wtypes.text, readonly=True)
    lessee_id = wsme.wsattr(wtypes.text)
    lessee = wsme.wsattr(wtypes.text, readonly=True)
    resource_type = wsme.wsattr(wtypes.text)
    resource_uuid = wsme.wsattr(wtypes.text, mandatory=True)
    resource = wsme.wsattr(wtypes.text, readonly=True)
    resource_class = wsme.wsattr(wtypes.text)
    start_time = wsme.wsattr(datetime.datetime)
    end_time = wsme.wsattr(datetime.datetime)
    status = wsme.wsattr(wtypes.text, readonly=True)
    properties = {wtypes.text: types.jsontype}
    availabilities = wsme.wsattr([[datetime.datetime]], readonly=True)
    parent_lease_uuid = wsme.wsattr(wtypes.text, readonly=True)

    def __init__(self, **kwargs):

        self.fields = offer_obj.Offer.fields
        for field in self.fields:
            setattr(self, field, kwargs.get(field, wtypes.Unset))

        for attr in ('availabilities', 'project', 'lessee',
                     'resource', 'resource_class'):
            setattr(self, attr, kwargs.get(attr, wtypes.Unset))


class OfferCollection(types.Collection):
    offers = [Offer]

    def __init__(self, **kwargs):
        self._type = 'offers'


class OffersController(rest.RestController):

    _custom_actions = {
        'claim': ['POST']
    }

    @wsme_pecan.wsexpose(Offer, wtypes.text)
    def get_one(self, offer_id):
        request = pecan.request.context
        cdict = request.to_policy_values()

        offer = utils.check_offer_policy_and_retrieve(
            request, 'esi_leap:offer:get', offer_id)
        utils.check_offer_lessee(cdict, offer)

        o = utils.offer_get_dict_with_added_info(offer)

        return Offer(**o)

    @wsme_pecan.wsexpose(OfferCollection, wtypes.text, wtypes.text,
                         wtypes.text, wtypes.text, datetime.datetime,
                         datetime.datetime, datetime.datetime,
                         datetime.datetime, wtypes.text)
    def get_all(self, project_id=None, resource_type=None,
                resource_class=None, resource_uuid=None,
                start_time=None, end_time=None,
                available_start_time=None, available_end_time=None,
                status=None):
        request = pecan.request.context
        cdict = request.to_policy_values()
        utils.policy_authorize('esi_leap:offer:get_all', cdict, cdict)

        if project_id is not None:
            project_id = keystone.get_project_uuid_from_ident(project_id)

        if resource_uuid is not None:
            if resource_type is None:
                resource_type = CONF.api.default_resource_type
            resource = get_resource_object(resource_type, resource_uuid)
            resource_uuid = resource.get_uuid()

        # either both are defined or both are None
        if bool(start_time) != bool(end_time):
            raise exception.InvalidTimeAPICommand(resource='an offer',
                                                  start_time=str(start_time),
                                                  end_time=str(end_time))
        if (start_time or end_time) and (end_time <= start_time):
            raise exception.InvalidTimeAPICommand(resource='an offer',
                                                  start_time=str(start_time),
                                                  end_time=str(end_time))

        if bool(available_start_time) != bool(available_end_time):
            raise exception.InvalidAvailabilityAPICommand(
                a_start=str(available_start_time),
                a_end=str(available_end_time))
        if ((available_start_time or available_end_time) and
                (available_end_time <= available_start_time)):
            raise exception.InvalidAvailabilityAPICommand(
                a_start=str(available_start_time),
                a_end=str(available_end_time))

        if status is None:
            status = statuses.OFFER_CAN_DELETE
        elif status == 'any':
            status = None
        else:
            status = [status]

        try:
            utils.policy_authorize('esi_leap:offer:offer_admin', cdict, cdict)
            lessee_id = None
        except exception.HTTPForbidden:
            lessee_id = cdict['project_id']

        filters = {
            'project_id': project_id,
            'lessee_id': lessee_id,
            'resource_type': resource_type,
            'resource_uuid': resource_uuid,
            'status': status,
            'start_time': start_time,
            'end_time': end_time,
            'available_start_time': available_start_time,
            'available_end_time': available_end_time,
        }

        # unpack iterator to tuple so we can use 'del'
        for k, v in tuple(filters.items()):
            if v is None:
                del filters[k]

        offer_collection = OfferCollection()
        offers = offer_obj.Offer.get_all(filters, request)

        offer_collection.offers = []

        if len(offers) > 0:
            project_list = None
            node_list = None
            with concurrent.futures.ThreadPoolExecutor() as executor:
                f1 = executor.submit(ironic.get_node_list)
                f2 = executor.submit(keystone.get_project_list)
                node_list = f1.result()
                project_list = f2.result()

            offers_with_added_info = [
                Offer(**utils.offer_get_dict_with_added_info(o, project_list,
                                                             node_list))
                for o in offers]
            if resource_class:
                offer_collection.offers = [
                    o for o in offers_with_added_info
                    if o.resource_class == resource_class]
            else:
                offer_collection.offers = offers_with_added_info

        return offer_collection

    @wsme_pecan.wsexpose(Offer, body=Offer, status_code=http_client.CREATED)
    def post(self, new_offer):
        request = pecan.request.context
        cdict = request.to_policy_values()
        utils.policy_authorize('esi_leap:offer:create', cdict, cdict)

        offer_dict = new_offer.to_dict()
        offer_dict['project_id'] = request.project_id
        offer_dict['uuid'] = uuidutils.generate_uuid()
        if 'resource_type' not in offer_dict:
            offer_dict['resource_type'] = CONF.api.default_resource_type
        resource = get_resource_object(offer_dict['resource_type'],
                                       offer_dict['resource_uuid'])
        offer_dict['resource_uuid'] = resource.get_uuid()

        if 'lessee_id' in offer_dict:
            offer_dict['lessee_id'] = keystone.get_project_uuid_from_ident(
                offer_dict['lessee_id'])

        if 'start_time' not in offer_dict:
            offer_dict['start_time'] = datetime.datetime.now()
        if 'end_time' not in offer_dict:
            offer_dict['end_time'] = datetime.datetime.max

        if offer_dict['start_time'] >= offer_dict['end_time']:
            raise exception.InvalidTimeRange(
                resource='an offer',
                start_time=str(offer_dict['start_time']),
                end_time=str(offer_dict['end_time']))

        try:
            utils.check_resource_admin(cdict, resource, request.project_id)
        except exception.HTTPResourceForbidden:
            parent_lease_uuid = utils.check_resource_lease_admin(
                cdict,
                resource,
                request.project_id,
                offer_dict.get('start_time'),
                offer_dict.get('end_time'))
            if parent_lease_uuid is None:
                raise
            offer_dict['parent_lease_uuid'] = parent_lease_uuid

        o = offer_obj.Offer(**offer_dict)
        o.create()
        return Offer(**utils.offer_get_dict_with_added_info(o))

    @wsme_pecan.wsexpose(Offer, wtypes.text)
    def delete(self, offer_id):
        request = pecan.request.context

        offer = utils.check_offer_policy_and_retrieve(
            request, 'esi_leap:offer:delete', offer_id,
            statuses.OFFER_CAN_DELETE)
        offer.cancel()

    @wsme_pecan.wsexpose(lease.Lease, wtypes.text, body=lease.Lease,
                         status_code=http_client.CREATED)
    def claim(self, offer_uuid, new_lease):
        request = pecan.request.context
        cdict = request.to_policy_values()

        offer = utils.check_offer_policy_and_retrieve(
            request, 'esi_leap:offer:claim', offer_uuid, [statuses.AVAILABLE])
        utils.check_offer_lessee(cdict, offer)

        lease_dict = new_lease.to_dict()
        lease_dict['project_id'] = request.project_id
        lease_dict['uuid'] = uuidutils.generate_uuid()
        lease_dict['offer_uuid'] = offer_uuid
        lease_dict['resource_type'] = offer.resource_type
        lease_dict['resource_uuid'] = offer.resource_uuid
        lease_dict['owner_id'] = offer.project_id

        if offer.parent_lease_uuid is not None:
            lease_dict['parent_lease_uuid'] = offer.parent_lease_uuid

        if 'start_time' not in lease_dict:
            lease_dict['start_time'] = datetime.datetime.now()

        if 'end_time' not in lease_dict:
            q = offer.get_next_lease_start_time(
                lease_dict['start_time'])
            if q is None:
                lease_dict['end_time'] = offer.end_time
            else:
                lease_dict['end_time'] = q.start_time

        new_lease = lease_obj.Lease(**lease_dict)
        new_lease.create(request)
        return lease.Lease(**utils.lease_get_dict_with_added_info(new_lease))
